import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm
    import sqlalchemy.orm.collections


@cassiopeia.type.core.common.inheritdocs
class MatchDetail(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        mapId (int): match map ID
        matchCreation (int): match creation time. Designates when the team select lobby is created and/or the match is made through match making, not when the game actually starts.
        matchDuration (int): match duration
        matchId (int): ID of the match
        matchMode (str): match mode (Legal values: CLASSIC, ODIN, ARAM, TUTORIAL, ONEFORALL, ASCENSION, FIRSTBLOOD, KINGPORO)
        matchType (str): match type (Legal values: CUSTOM_GAME, MATCHED_GAME, TUTORIAL_GAME)
        matchVersion (str): match version
        participantIdentities (list<ParticipantIdentity>): participant identity information
        participants (list<Participant>): participant information
        platformId (str): platform ID of the match
        queueType (str): match queue type (Legal values: CUSTOM, NORMAL_5x5_BLIND, RANKED_SOLO_5x5, RANKED_PREMADE_5x5, BOT_5x5, NORMAL_3x3, RANKED_PREMADE_3x3, NORMAL_5x5_DRAFT, ODIN_5x5_BLIND, ODIN_5x5_DRAFT, BOT_ODIN_5x5, BOT_5x5_INTRO, BOT_5x5_BEGINNER, BOT_5x5_INTERMEDIATE, RANKED_TEAM_3x3, RANKED_TEAM_5x5, BOT_TT_3x3, GROUP_FINDER_5x5, ARAM_5x5, ONEFORALL_5x5, FIRSTBLOOD_1x1, FIRSTBLOOD_2x2, SR_6x6, URF_5x5, ONEFORALL_MIRRORMODE_5x5, BOT_URF_5x5, NIGHTMARE_BOT_5x5_RANK1, NIGHTMARE_BOT_5x5_RANK2, NIGHTMARE_BOT_5x5_RANK5, ASCENSION_5x5, HEXAKILL, KING_PORO_5x5, COUNTER_PICK)
        region (str): region where the match was played
        season (str): season match was played (Legal values: PRESEASON3, SEASON3, PRESEASON2014, SEASON2014, PRESEASON2015, SEASON2015)
        teams (list<Team>): team information
        timeline (Timeline): match timeline data (not included by default)
    """
    def __init__(self, dictionary):
        self.mapId = dictionary.get("mapId", 0)
        self.matchCreation = dictionary.get("matchCreation", 0)
        self.matchDuration = dictionary.get("matchDuration", 0)
        self.matchId = dictionary.get("matchId", 0)
        self.matchMode = dictionary.get("matchMode", "")
        self.matchType = dictionary.get("matchType", "")
        self.matchVersion = dictionary.get("matchVersion", "")
        self.participantIdentities = [(ParticipantIdentity(pi) if not isinstance(pi, ParticipantIdentity) else pi) for pi in dictionary.get("participantIdentities", []) if pi]
        self.participants = [(Participant(p) if not isinstance(p, Participant) else p) for p in dictionary.get("participants", []) if p]
        self.platformId = dictionary.get("platformId", "")
        self.queueType = dictionary.get("queueType", "")
        self.region = dictionary.get("region", "")
        self.season = dictionary.get("season", "")
        self.teams = [(Team(t) if not isinstance(t, Team) else t) for t in dictionary.get("teams", []) if t]
        val = dictionary.get("timeline", None)
        self.timeline = Timeline(val) if val and not isinstance(val, Timeline) else val

    @property
    def item_ids(self):
        """
        Gets all item IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            s = p.stats
            if s.item0:
                ids.add(s.item0)
            if s.item1:
                ids.add(s.item1)
            if s.item2:
                ids.add(s.item2)
            if s.item3:
                ids.add(s.item3)
            if s.item4:
                ids.add(s.item4)
            if s.item5:
                ids.add(s.item5)
            if s.item6:
                ids.add(s.item6)
        return ids

    @property
    def champion_ids(self):
        """
        Gets all item IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            if p.championId:
                ids.add(p.championId)
        for t in self.teams:
            for b in t.bans:
                if b.championId:
                    ids.add(b.championId)
        return ids

    @property
    def mastery_ids(self):
        """
        Gets all item IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            for m in p.masteries:
                if m.masteryId:
                    ids.add(m.masteryId)
        return ids

    @property
    def rune_ids(self):
        """
        Gets all item IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            for r in p.runes:
                if r.runeId:
                    ids.add(r.runeId)
        return ids

    @property
    def summoner_ids(self):
        """
        Gets all item IDs contained in this object
        """
        ids = set()
        for p in self.participantIdentities:
            if p.player and p.player.summonerId:
                ids.add(p.player.summonerId)
        return ids

    @property
    def summoner_spell_ids(self):
        """
        Gets all item IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            if p.spell1Id:
                ids.add(p.spell1Id)
            if p.spell2Id:
                ids.add(p.spell2Id)
        return ids


@cassiopeia.type.core.common.inheritdocs
class Participant(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all item IDs contained in this object
    """
    def __init__(self, dictionary):
        self.championId = dictionary.get("championId", 0)
        self.highestAchievedSeasonTier = dictionary.get("highestAchievedSeasonTier", "")
        self.masteries = [(Mastery(m) if not isinstance(m, Mastery) else m) for m in dictionary.get("masteries", []) if m]
        self.participantId = dictionary.get("participantId", 0)
        self.runes = [(Rune(r) if not isinstance(r, Rune) else r) for r in dictionary.get("runes", []) if r]
        self.spell1Id = dictionary.get("spell1Id", 0)
        self.spell2Id = dictionary.get("spell2Id", 0)
        val = dictionary.get("stats", None)
        self.stats = ParticipantStats(val) if val and not isinstance(val, ParticipantStats) else val
        self.teamId = dictionary.get("teamId", 0)
        val = dictionary.get("timeline", None)
        self.timeline = ParticipantTimeline(val) if val and not isinstance(val, ParticipantTimeline) else val


@cassiopeia.type.core.common.inheritdocs
class ParticipantIdentity(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all champion IDs contained in this object
    """
    def __init__(self, dictionary):
        self.participantId = dictionary.get("participantId", 0)
        val = dictionary.get("player", None)
        self.player = Player(val) if val and not isinstance(val, Player) else val


@cassiopeia.type.core.common.inheritdocs
class Team(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all mastery IDs contained in this object
    """
    def __init__(self, dictionary):
        self.bans = [(BannedChampion(c) if not isinstance(c, BannedChampion) else c) for c in dictionary.get("bans", []) if c]
        self.baronKills = dictionary.get("baronKills", 0)
        self.dominionVictoryScore = dictionary.get("dominionVictoryScore", 0)
        self.dragonKills = dictionary.get("dragonKills", 0)
        self.firstBaron = dictionary.get("firstBaron", False)
        self.firstBlood = dictionary.get("firstBlood", False)
        self.firstDragon = dictionary.get("firstDragon", False)
        self.firstInhibitor = dictionary.get("firstInhibitor", False)
        self.firstRiftHerald = dictionary.get("firstRiftHerald", False)
        self.firstTower = dictionary.get("firstTower", False)
        self.inhibitorKills = dictionary.get("inhibitorKills", 0)
        self.riftHeraldKills = dictionary.get("riftHeraldKills", 0)
        self.teamId = dictionary.get("teamId", 0)
        self.towerKills = dictionary.get("towerKills", 0)
        self.vilemawKills = dictionary.get("vilemawKills", 0)
        self.winner = dictionary.get("winner", False)


@cassiopeia.type.core.common.inheritdocs
class Timeline(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all rune IDs contained in this object
    """
    def __init__(self, dictionary):
        self.frameInterval = dictionary.get("frameInterval", 0)
        self.frames = [(Frame(f) if not isinstance(f, Frame) else f) for f in dictionary.get("frames", []) if f]


@cassiopeia.type.core.common.inheritdocs
class Mastery(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all summoner IDs contained in this object
    """
    def __init__(self, dictionary):
        self.masteryId = dictionary.get("masteryId", 0)
        self.rank = dictionary.get("rank", 0)


@cassiopeia.type.core.common.inheritdocs
class ParticipantStats(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all summoner spell IDs contained in this object
    """
    def __init__(self, dictionary):
        self.assists = dictionary.get("assists", 0)
        self.champLevel = dictionary.get("champLevel", 0)
        self.combatPlayerScore = dictionary.get("combatPlayerScore", 0)
        self.deaths = dictionary.get("deaths", 0)
        self.doubleKills = dictionary.get("doubleKills", 0)
        self.firstBloodAssist = dictionary.get("firstBloodAssist", False)
        self.firstBloodKill = dictionary.get("firstBloodKill", False)
        self.firstInhibitorAssist = dictionary.get("firstInhibitorAssist", False)
        self.firstInhibitorKill = dictionary.get("firstInhibitorKill", False)
        self.firstTowerAssist = dictionary.get("firstTowerAssist", False)
        self.firstTowerKill = dictionary.get("firstTowerKill", False)
        self.goldEarned = dictionary.get("goldEarned", 0)
        self.goldSpent = dictionary.get("goldSpent", 0)
        self.inhibitorKills = dictionary.get("inhibitorKills", 0)
        self.item0 = dictionary.get("item0", 0)
        self.item1 = dictionary.get("item1", 0)
        self.item2 = dictionary.get("item2", 0)
        self.item3 = dictionary.get("item3", 0)
        self.item4 = dictionary.get("item4", 0)
        self.item5 = dictionary.get("item5", 0)
        self.item6 = dictionary.get("item6", 0)
        self.killingSprees = dictionary.get("killingSprees", 0)
        self.kills = dictionary.get("kills", 0)
        self.largestCriticalStrike = dictionary.get("largestCriticalStrike", 0)
        self.largestKillingSpree = dictionary.get("largestKillingSpree", 0)
        self.largestMultiKill = dictionary.get("largestMultiKill", 0)
        self.magicDamageDealt = dictionary.get("magicDamageDealt", 0)
        self.magicDamageDealtToChampions = dictionary.get("magicDamageDealtToChampions", 0)
        self.magicDamageTaken = dictionary.get("magicDamageTaken", 0)
        self.minionsKilled = dictionary.get("minionsKilled", 0)
        self.neutralMinionsKilled = dictionary.get("neutralMinionsKilled", 0)
        self.neutralMinionsKilledEnemyJungle = dictionary.get("neutralMinionsKilledEnemyJungle", 0)
        self.neutralMinionsKilledTeamJungle = dictionary.get("neutralMinionsKilledTeamJungle", 0)
        self.nodeCapture = dictionary.get("nodeCapture", 0)
        self.nodeCaptureAssist = dictionary.get("nodeCaptureAssist", 0)
        self.nodeNeutralize = dictionary.get("nodeNeutralize", 0)
        self.nodeNeutralizeAssist = dictionary.get("nodeNeutralizeAssist", 0)
        self.objectivePlayerScore = dictionary.get("objectivePlayerScore", 0)
        self.pentaKills = dictionary.get("pentaKills", 0)
        self.physicalDamageDealt = dictionary.get("physicalDamageDealt", 0)
        self.physicalDamageDealtToChampions = dictionary.get("physicalDamageDealtToChampions", 0)
        self.physicalDamageTaken = dictionary.get("physicalDamageTaken", 0)
        self.quadraKills = dictionary.get("quadraKills", 0)
        self.sightWardsBoughtInGame = dictionary.get("sightWardsBoughtInGame", 0)
        self.teamObjective = dictionary.get("teamObjective", 0)
        self.totalDamageDealt = dictionary.get("totalDamageDealt", 0)
        self.totalDamageDealtToChampions = dictionary.get("totalDamageDealtToChampions", 0)
        self.totalDamageTaken = dictionary.get("totalDamageTaken", 0)
        self.totalHeal = dictionary.get("totalHeal", 0)
        self.totalPlayerScore = dictionary.get("totalPlayerScore", 0)
        self.totalScoreRank = dictionary.get("totalScoreRank", 0)
        self.totalTimeCrowdControlDealt = dictionary.get("totalTimeCrowdControlDealt", 0)
        self.totalUnitsHealed = dictionary.get("totalUnitsHealed", 0)
        self.towerKills = dictionary.get("towerKills", 0)
        self.tripleKills = dictionary.get("tripleKills", 0)
        self.trueDamageDealt = dictionary.get("trueDamageDealt", 0)
        self.trueDamageDealtToChampions = dictionary.get("trueDamageDealtToChampions", 0)
        self.trueDamageTaken = dictionary.get("trueDamageTaken", 0)
        self.unrealKills = dictionary.get("unrealKills", 0)
        self.visionWardsBoughtInGame = dictionary.get("visionWardsBoughtInGame", 0)
        self.wardsKilled = dictionary.get("wardsKilled", 0)
        self.wardsPlaced = dictionary.get("wardsPlaced", 0)
        self.winner = dictionary.get("winner", False)


@cassiopeia.type.core.common.inheritdocs
class ParticipantTimeline(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        championId (int): champion ID
        highestAchievedSeasonTier (str): highest ranked tier achieved for the previous season, if any, otherwise null. Used to display border in game loading screen. (Legal values: CHALLENGER, MASTER, DIAMOND, PLATINUM, GOLD, SILVER, BRONZE, UNRANKED)
        masteries (list<Mastery>): list of mastery information
        participantId (int): participant ID
        runes (list<Rune>): list of rune information
        spell1Id (int): first summoner spell ID
        spell2Id (int): second summoner spell ID
        stats (ParticipantStats): participant statistics
        teamId (int): team ID
        timeline (ParticipantTimeline): timeline data. Delta fields refer to values for the specified period (e.g., the gold per minute over the first 10 minutes of the game versus the second 20 minutes of the game. Diffs fields refer to the deltas versus the calculated lane opponent(s).
    """
    def __init__(self, dictionary):
        val = dictionary.get("ancientGolemAssistsPerMinCounts", None)
        self.ancientGolemAssistsPerMinCounts = ParticipantTimelineData(val, "ancientGolemAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("ancientGolemKillsPerMinCounts", None)
        self.ancientGolemKillsPerMinCounts = ParticipantTimelineData(val, "ancientGolemKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("assistedLaneDeathsPerMinDeltas", None)
        self.assistedLaneDeathsPerMinDeltas = ParticipantTimelineData(val, "assistedLaneDeathsPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("assistedLaneKillsPerMinDeltas", None)
        self.assistedLaneKillsPerMinDeltas = ParticipantTimelineData(val, "assistedLaneKillsPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("baronAssistsPerMinCounts", None)
        self.baronAssistsPerMinCounts = ParticipantTimelineData(val, "baronAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("baronKillsPerMinCounts", None)
        self.baronKillsPerMinCounts = ParticipantTimelineData(val, "baronKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("creepsPerMinDeltas", None)
        self.creepsPerMinDeltas = ParticipantTimelineData(val, "creepsPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("csDiffPerMinDeltas", None)
        self.csDiffPerMinDeltas = ParticipantTimelineData(val, "csDiffPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("damageTakenDiffPerMinDeltas", None)
        self.damageTakenDiffPerMinDeltas = ParticipantTimelineData(val, "damageTakenDiffPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("damageTakenPerMinDeltas", None)
        self.damageTakenPerMinDeltas = ParticipantTimelineData(val, "damageTakenPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("dragonAssistsPerMinCounts", None)
        self.dragonAssistsPerMinCounts = ParticipantTimelineData(val, "dragonAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("dragonKillsPerMinCounts", None)
        self.dragonKillsPerMinCounts = ParticipantTimelineData(val, "dragonKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("elderLizardAssistsPerMinCounts", None)
        self.elderLizardAssistsPerMinCounts = ParticipantTimelineData(val, "elderLizardAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("elderLizardKillsPerMinCounts", None)
        self.elderLizardKillsPerMinCounts = ParticipantTimelineData(val, "elderLizardKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("goldPerMinDeltas", None)
        self.goldPerMinDeltas = ParticipantTimelineData(val, "goldPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("inhibitorAssistsPerMinCounts", None)
        self.inhibitorAssistsPerMinCounts = ParticipantTimelineData(val, "inhibitorAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("inhibitorKillsPerMinCounts", None)
        self.inhibitorKillsPerMinCounts = ParticipantTimelineData(val, "inhibitorKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        self.lane = dictionary.get("lane", "")
        self.role = dictionary.get("role", "")
        val = dictionary.get("towerAssistsPerMinCounts", None)
        self.towerAssistsPerMinCounts = ParticipantTimelineData(val, "towerAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("towerKillsPerMinCounts", None)
        self.towerKillsPerMinCounts = ParticipantTimelineData(val, "towerKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("towerKillsPerMinDeltas", None)
        self.towerKillsPerMinDeltas = ParticipantTimelineData(val, "towerKillsPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("vilemawAssistsPerMinCounts", None)
        self.vilemawAssistsPerMinCounts = ParticipantTimelineData(val, "vilemawAssistsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("vilemawKillsPerMinCounts", None)
        self.vilemawKillsPerMinCounts = ParticipantTimelineData(val, "vilemawKillsPerMinCounts") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("wardsPerMinDeltas", None)
        self.wardsPerMinDeltas = ParticipantTimelineData(val, "wardsPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("xpDiffPerMinDeltas", None)
        self.xpDiffPerMinDeltas = ParticipantTimelineData(val, "xpDiffPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val
        val = dictionary.get("xpPerMinDeltas", None)
        self.xpPerMinDeltas = ParticipantTimelineData(val, "xpPerMinDeltas") if val and not isinstance(val, ParticipantTimelineData) else val


@cassiopeia.type.core.common.inheritdocs
class Rune(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        participantId (int): participant ID
        player (Player): player information
    """
    def __init__(self, dictionary):
        self.rank = dictionary.get("rank", 0)
        self.runeId = dictionary.get("runeId", 0)


@cassiopeia.type.core.common.inheritdocs
class Player(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        bans (list<BannedChampion>): if game was draft mode, contains banned champion data, otherwise null
        baronKills (int): number of times the team killed baron
        dominionVictoryScore (int): if game was a dominion game, specifies the points the team had at game end, otherwise null
        dragonKills (int): number of times the team killed dragon
        firstBaron (bool): flag indicating whether or not the team got the first baron kill
        firstBlood (bool): flag indicating whether or not the team got first blood
        firstDragon (bool): flag indicating whether or not the team got the first dragon kill
        firstInhibitor (bool): flag indicating whether or not the team destroyed the first inhibitor
        firstRiftHerald (bool): flag indicating whether or not the team got the first rift herald kill
        firstTower (bool): flag indicating whether or not the team destroyed the first tower
        inhibitorKills (int): number of inhibitors the team destroyed
        riftHeraldKills (int): number of times the team killed rift herald
        teamId (int): team ID
        towerKills (int): number of towers the team destroyed
        vilemawKills (int): number of times the team killed vilemaw
        winner (bool): flag indicating whether or not the team won
    """
    def __init__(self, dictionary):
        self.matchHistoryUri = dictionary.get("matchHistoryUri", "")
        self.profileIcon = dictionary.get("profileIcon", 0)
        self.summonerId = dictionary.get("summonerId", 0)
        self.summonerName = dictionary.get("summonerName", "")


@cassiopeia.type.core.common.inheritdocs
class BannedChampion(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        frameInterval (int): time between each returned frame in milliseconds
        frames (list<Frame>): list of timeline frames for the game
    """
    def __init__(self, dictionary):
        self.championId = dictionary.get("championId", 0)
        self.pickTurn = dictionary.get("pickTurn", 0)


@cassiopeia.type.core.common.inheritdocs
class Frame(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        masteryId (int): mastery ID
        rank (int): mastery rank
    """
    def __init__(self, dictionary):
        self.events = [(Event(e) if not isinstance(e, Event) else e) for e in dictionary.get("events", []) if e]
        self.participantFrames = {i: ParticipantFrame(pf) if not isinstance(pf, ParticipantFrame) else pf for i, pf in dictionary.get("participantFrames", {}).items()}
        self.timestamp = dictionary.get("timestamp", 0)


@cassiopeia.type.core.common.inheritdocs
class ParticipantTimelineData(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        assists (int): number of assists
        champLevel (int): champion level achieved
        combatPlayerScore (int): if game was a dominion game, player's combat score, otherwise 0
        deaths (int): number of deaths
        doubleKills (int): number of double kills
        firstBloodAssist (bool): flag indicating if participant got an assist on first blood
        firstBloodKill (bool): flag indicating if participant got first blood
        firstInhibitorAssist (bool): flag indicating if participant got an assist on the first inhibitor
        firstInhibitorKill (bool): flag indicating if participant destroyed the first inhibitor
        firstTowerAssist (bool): flag indicating if participant got an assist on the first tower
        firstTowerKill (bool): flag indicating if participant destroyed the first tower
        goldEarned (int): gold earned
        goldSpent (int): gold spent
        inhibitorKills (int): number of inhibitor kills
        item0 (int): frst item ID
        item1 (int): second item ID
        item2 (int): third item ID
        item3 (int): fourth item ID
        item4 (int): fifth item ID
        item5 (int): sixth item ID
        item6 (int): seventh item ID
        killingSprees (int): number of killing sprees
        kills (int): number of kills
        largestCriticalStrike (int): largest critical strike
        largestKillingSpree (int): largest killing spree
        largestMultiKill (int): largest multi kill
        magicDamageDealt (int): magical damage dealt
        magicDamageDealtToChampions (int): magical damage dealt to champions
        magicDamageTaken (int): magic damage taken
        minionsKilled (int): minions killed
        neutralMinionsKilled (int): neutral minions killed
        neutralMinionsKilledEnemyJungle (int): neutral jungle minions killed in the enemy team's jungle
        neutralMinionsKilledTeamJungle (int): neutral jungle minions killed in your team's jungle
        nodeCapture (int): if game was a dominion game, number of node captures
        nodeCaptureAssist (int): if game was a dominion game, number of node capture assists
        nodeNeutralize (int): if game was a dominion game, number of node neutralizations
        nodeNeutralizeAssist (int): if game was a dominion game, number of node neutralization assists
        objectivePlayerScore (int): if game was a dominion game, player's objectives score, otherwise 0
        pentaKills (int): number of penta kills
        physicalDamageDealt (int): physical damage dealt
        physicalDamageDealtToChampions (int): physical damage dealt to champions
        physicalDamageTaken (int): physical damage taken
        quadraKills (int): number of quadra kills
        sightWardsBoughtInGame (int): sight wards purchased
        teamObjective (int): if game was a dominion game, number of completed team objectives (i.e., quests)
        totalDamageDealt (int): total damage dealt
        totalDamageDealtToChampions (int): total damage dealt to champions
        totalDamageTaken (int): total damage taken
        totalHeal (int): total heal amount
        totalPlayerScore (int): if game was a dominion game, player's total score, otherwise 0
        totalScoreRank (int): if game was a dominion game, team rank of the player's total score (e.g., 1-5)
        totalTimeCrowdControlDealt (int): total dealt crowd control time
        totalUnitsHealed (int): total units healed
        towerKills (int): number of tower kills
        tripleKills (int): number of triple kills
        trueDamageDealt (int): true damage dealt
        trueDamageDealtToChampions (int): true damage dealt to champions
        trueDamageTaken (int): true damage taken
        unrealKills (int): number of unreal kills
        visionWardsBoughtInGame (int): vision wards purchased
        wardsKilled (int): number of wards killed
        wardsPlaced (int): number of wards placed
        winner (bool): flag indicating whether or not the participant won
    """
    def __init__(self, dictionary, type_=None):
        self.tenToTwenty = dictionary.get("tenToTwenty", 0.0)
        self.thirtyToEnd = dictionary.get("thirtyToEnd", 0.0)
        self.twentyToThirty = dictionary.get("twentyToThirty", 0.0)
        self.zeroToTen = dictionary.get("zeroToTen", 0.0)
        self._type = type_


@cassiopeia.type.core.common.inheritdocs
class Event(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        ancientGolemAssistsPerMinCounts (ParticipantTimelineData): ancient golem assists per minute timeline counts
        ancientGolemKillsPerMinCounts (ParticipantTimelineData): ancient golem kills per minute timeline counts
        assistedLaneDeathsPerMinDeltas (ParticipantTimelineData): assisted lane deaths per minute timeline data
        assistedLaneKillsPerMinDeltas (ParticipantTimelineData): assisted lane kills per minute timeline data
        baronAssistsPerMinCounts (ParticipantTimelineData): baron assists per minute timeline counts
        baronKillsPerMinCounts (ParticipantTimelineData): baron kills per minute timeline counts
        creepsPerMinDeltas (ParticipantTimelineData): creeps per minute timeline data
        csDiffPerMinDeltas (ParticipantTimelineData): creep score difference per minute timeline data
        damageTakenDiffPerMinDeltas (ParticipantTimelineData): damage taken difference per minute timeline data
        damageTakenPerMinDeltas (ParticipantTimelineData): damage taken per minute timeline data
        dragonAssistsPerMinCounts (ParticipantTimelineData): dragon assists per minute timeline counts
        dragonKillsPerMinCounts (ParticipantTimelineData): dragon kills per minute timeline counts
        elderLizardAssistsPerMinCounts (ParticipantTimelineData): elder lizard assists per minute timeline counts
        elderLizardKillsPerMinCounts (ParticipantTimelineData): elder lizard kills per minute timeline counts
        goldPerMinDeltas (ParticipantTimelineData): gold per minute timeline data
        inhibitorAssistsPerMinCounts (ParticipantTimelineData): inhibitor assists per minute timeline counts
        inhibitorKillsPerMinCounts (ParticipantTimelineData): inhibitor kills per minute timeline counts
        lane (str): participant's lane (Legal values: MID, MIDDLE, TOP, JUNGLE, BOT, BOTTOM)
        role (str): participant's role (Legal values: DUO, NONE, SOLO, DUO_CARRY, DUO_SUPPORT)
        towerAssistsPerMinCounts (ParticipantTimelineData): tower assists per minute timeline counts
        towerKillsPerMinCounts (ParticipantTimelineData): tower kills per minute timeline counts
        towerKillsPerMinDeltas (ParticipantTimelineData): tower kills per minute timeline data
        vilemawAssistsPerMinCounts (ParticipantTimelineData): vilemaw assists per minute timeline counts
        vilemawKillsPerMinCounts (ParticipantTimelineData): vilemaw kills per minute timeline counts
        wardsPerMinDeltas (ParticipantTimelineData): wards placed per minute timeline data
        xpDiffPerMinDeltas (ParticipantTimelineData): experience difference per minute timeline data
        xpPerMinDeltas (ParticipantTimelineData): experience per minute timeline data
    """
    def __init__(self, dictionary):
        self.ascendedType = dictionary.get("ascendedType", "")
        self.assistingParticipantIds = dictionary.get("assistingParticipantIds", [])
        self.buildingType = dictionary.get("buildingType", "")
        self.creatorId = dictionary.get("creatorId", 0)
        self.eventType = dictionary.get("eventType", "")
        self.itemAfter = dictionary.get("itemAfter", 0)
        self.itemBefore = dictionary.get("itemBefore", 0)
        self.itemId = dictionary.get("itemId", 0)
        self.killerId = dictionary.get("killerId", 0)
        self.laneType = dictionary.get("laneType", "")
        self.levelUpType = dictionary.get("levelUpType", "")
        self.monsterType = dictionary.get("monsterType", "")
        self.participantId = dictionary.get("participantId", 0)
        self.pointCaptured = dictionary.get("pointCaptured", "")
        val = dictionary.get("position", None)
        self.position = Position(val) if val and not isinstance(val, Position) else val
        self.skillSlot = dictionary.get("skillSlot", 0)
        self.teamId = dictionary.get("teamId", 0)
        self.timestamp = dictionary.get("timestamp", 0)
        self.towerType = dictionary.get("towerType", "")
        self.victimId = dictionary.get("victimId", 0)
        self.wardType = dictionary.get("wardType", "")


@cassiopeia.type.core.common.inheritdocs
class ParticipantFrame(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        rank (int): rune rank
        runeId (int): rune ID
    """
    def __init__(self, dictionary):
        self.currentGold = dictionary.get("currentGold", 0)
        self.dominionScore = dictionary.get("dominionScore", 0)
        self.jungleMinionsKilled = dictionary.get("jungleMinionsKilled", 0)
        self.level = dictionary.get("level", 0)
        self.minionsKilled = dictionary.get("minionsKilled", 0)
        self.participantId = dictionary.get("participantId", 0)
        val = dictionary.get("position", None)
        self.position = Position(val) if val and not isinstance(val, Position) else val
        self.teamScore = dictionary.get("teamScore", 0)
        self.totalGold = dictionary.get("totalGold", 0)
        self.xp = dictionary.get("xp", 0)


@cassiopeia.type.core.common.inheritdocs
class Position(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        matchHistoryUri (str): match history URI
        profileIcon (int): profile icon ID
        summonerId (int): summoner ID
        summonerName (str): summoner name
    """
    def __init__(self, dictionary):
        self.x = dictionary.get("x", 0)
        self.y = dictionary.get("y", 0)


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_match_detail():
    global MatchDetail

    @cassiopeia.type.core.common.inheritdocs
    class MatchDetail(MatchDetail, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchDetail"
        mapId = sqlalchemy.Column(sqlalchemy.Integer)
        matchCreation = sqlalchemy.Column(sqlalchemy.BigInteger)
        matchDuration = sqlalchemy.Column(sqlalchemy.Integer)
        matchId = sqlalchemy.Column(sqlalchemy.BigInteger, primary_key=True)
        matchMode = sqlalchemy.Column(sqlalchemy.String(30))
        matchType = sqlalchemy.Column(sqlalchemy.String(30))
        matchVersion = sqlalchemy.Column(sqlalchemy.String(30))
        participantIdentities = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantIdentity", cascade="all, delete-orphan", passive_deletes=True)
        participants = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Participant", cascade="all, delete-orphan", passive_deletes=True)
        platformId = sqlalchemy.Column(sqlalchemy.String(30))
        queueType = sqlalchemy.Column(sqlalchemy.String(30))
        region = sqlalchemy.Column(sqlalchemy.String(30))
        season = sqlalchemy.Column(sqlalchemy.String(30))
        teams = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Team", cascade="all, delete-orphan", passive_deletes=True)
        timeline = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Timeline", uselist=False, cascade="all, delete-orphan", passive_deletes=True)


def _sa_bind_participant():
    global Participant

    @cassiopeia.type.core.common.inheritdocs
    class Participant(Participant, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchParticipant"
        championId = sqlalchemy.Column(sqlalchemy.Integer)
        highestAchievedSeasonTier = sqlalchemy.Column(sqlalchemy.String(30))
        masteries = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Mastery", cascade="all, delete-orphan", passive_deletes=True)
        participantId = sqlalchemy.Column(sqlalchemy.Integer)
        runes = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Rune", cascade="all, delete-orphan", passive_deletes=True)
        spell1Id = sqlalchemy.Column(sqlalchemy.Integer)
        spell2Id = sqlalchemy.Column(sqlalchemy.Integer)
        stats = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantStats", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        teamId = sqlalchemy.Column(sqlalchemy.Integer)
        timeline = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimeline", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _match_id = sqlalchemy.Column(sqlalchemy.BigInteger, sqlalchemy.ForeignKey("MatchDetail.matchId", ondelete="CASCADE"))


def _sa_bind_participant_identity():
    global ParticipantIdentity

    @cassiopeia.type.core.common.inheritdocs
    class ParticipantIdentity(ParticipantIdentity, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchParticipantIdentity"
        participantId = sqlalchemy.Column(sqlalchemy.Integer)
        player = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Player", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _match_id = sqlalchemy.Column(sqlalchemy.BigInteger, sqlalchemy.ForeignKey("MatchDetail.matchId", ondelete="CASCADE"))


def _sa_bind_team():
    global Team

    @cassiopeia.type.core.common.inheritdocs
    class Team(Team, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchTeam"
        bans = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.BannedChampion", cascade="all, delete-orphan", passive_deletes=True)
        baronKills = sqlalchemy.Column(sqlalchemy.Integer)
        dominionVictoryScore = sqlalchemy.Column(sqlalchemy.Integer)
        dragonKills = sqlalchemy.Column(sqlalchemy.Integer)
        firstBaron = sqlalchemy.Column(sqlalchemy.Boolean)
        firstBlood = sqlalchemy.Column(sqlalchemy.Boolean)
        firstDragon = sqlalchemy.Column(sqlalchemy.Boolean)
        firstInhibitor = sqlalchemy.Column(sqlalchemy.Boolean)
        firstTower = sqlalchemy.Column(sqlalchemy.Boolean)
        inhibitorKills = sqlalchemy.Column(sqlalchemy.Integer)
        teamId = sqlalchemy.Column(sqlalchemy.Integer)
        towerKills = sqlalchemy.Column(sqlalchemy.Integer)
        vilemawKills = sqlalchemy.Column(sqlalchemy.Integer)
        winner = sqlalchemy.Column(sqlalchemy.Boolean)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _match_id = sqlalchemy.Column(sqlalchemy.BigInteger, sqlalchemy.ForeignKey("MatchDetail.matchId", ondelete="CASCADE"))


def _sa_bind_timeline():
    global Timeline

    @cassiopeia.type.core.common.inheritdocs
    class Timeline(Timeline, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchTimeline"
        frameInterval = sqlalchemy.Column(sqlalchemy.Integer)
        frames = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Frame", cascade="all, delete-orphan", passive_deletes=True)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _match_id = sqlalchemy.Column(sqlalchemy.BigInteger, sqlalchemy.ForeignKey("MatchDetail.matchId", ondelete="CASCADE"))


def _sa_bind_mastery():
    global Mastery

    @cassiopeia.type.core.common.inheritdocs
    class Mastery(Mastery, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchMastery"
        masteryId = sqlalchemy.Column(sqlalchemy.Integer)
        rank = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipant._id", ondelete="CASCADE"))


def _sa_bind_participant_stats():
    global ParticipantStats

    @cassiopeia.type.core.common.inheritdocs
    class ParticipantStats(ParticipantStats, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchParticipantStats"
        assists = sqlalchemy.Column(sqlalchemy.Integer)
        champLevel = sqlalchemy.Column(sqlalchemy.Integer)
        combatPlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        deaths = sqlalchemy.Column(sqlalchemy.Integer)
        doubleKills = sqlalchemy.Column(sqlalchemy.Integer)
        firstBloodAssist = sqlalchemy.Column(sqlalchemy.Boolean)
        firstBloodKill = sqlalchemy.Column(sqlalchemy.Boolean)
        firstInhibitorAssist = sqlalchemy.Column(sqlalchemy.Boolean)
        firstInhibitorKill = sqlalchemy.Column(sqlalchemy.Boolean)
        firstTowerAssist = sqlalchemy.Column(sqlalchemy.Boolean)
        firstTowerKill = sqlalchemy.Column(sqlalchemy.Boolean)
        goldEarned = sqlalchemy.Column(sqlalchemy.Integer)
        goldSpent = sqlalchemy.Column(sqlalchemy.Integer)
        inhibitorKills = sqlalchemy.Column(sqlalchemy.Integer)
        item0 = sqlalchemy.Column(sqlalchemy.Integer)
        item1 = sqlalchemy.Column(sqlalchemy.Integer)
        item2 = sqlalchemy.Column(sqlalchemy.Integer)
        item3 = sqlalchemy.Column(sqlalchemy.Integer)
        item4 = sqlalchemy.Column(sqlalchemy.Integer)
        item5 = sqlalchemy.Column(sqlalchemy.Integer)
        item6 = sqlalchemy.Column(sqlalchemy.Integer)
        killingSprees = sqlalchemy.Column(sqlalchemy.Integer)
        kills = sqlalchemy.Column(sqlalchemy.Integer)
        largestCriticalStrike = sqlalchemy.Column(sqlalchemy.Integer)
        largestKillingSpree = sqlalchemy.Column(sqlalchemy.Integer)
        largestMultiKill = sqlalchemy.Column(sqlalchemy.Integer)
        magicDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        magicDamageDealtToChampions = sqlalchemy.Column(sqlalchemy.Integer)
        magicDamageTaken = sqlalchemy.Column(sqlalchemy.Integer)
        minionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        neutralMinionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        neutralMinionsKilledEnemyJungle = sqlalchemy.Column(sqlalchemy.Integer)
        neutralMinionsKilledTeamJungle = sqlalchemy.Column(sqlalchemy.Integer)
        nodeCapture = sqlalchemy.Column(sqlalchemy.Integer)
        nodeCaptureAssist = sqlalchemy.Column(sqlalchemy.Integer)
        nodeNeutralize = sqlalchemy.Column(sqlalchemy.Integer)
        nodeNeutralizeAssist = sqlalchemy.Column(sqlalchemy.Integer)
        objectivePlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        pentaKills = sqlalchemy.Column(sqlalchemy.Integer)
        physicalDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        physicalDamageDealtToChampions = sqlalchemy.Column(sqlalchemy.Integer)
        physicalDamageTaken = sqlalchemy.Column(sqlalchemy.Integer)
        quadraKills = sqlalchemy.Column(sqlalchemy.Integer)
        sightWardsBoughtInGame = sqlalchemy.Column(sqlalchemy.Integer)
        teamObjective = sqlalchemy.Column(sqlalchemy.Integer)
        totalDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        totalDamageDealtToChampions = sqlalchemy.Column(sqlalchemy.Integer)
        totalDamageTaken = sqlalchemy.Column(sqlalchemy.Integer)
        totalHeal = sqlalchemy.Column(sqlalchemy.Integer)
        totalPlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        totalScoreRank = sqlalchemy.Column(sqlalchemy.Integer)
        totalTimeCrowdControlDealt = sqlalchemy.Column(sqlalchemy.Integer)
        totalUnitsHealed = sqlalchemy.Column(sqlalchemy.Integer)
        towerKills = sqlalchemy.Column(sqlalchemy.Integer)
        tripleKills = sqlalchemy.Column(sqlalchemy.Integer)
        trueDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        trueDamageDealtToChampions = sqlalchemy.Column(sqlalchemy.Integer)
        trueDamageTaken = sqlalchemy.Column(sqlalchemy.Integer)
        unrealKills = sqlalchemy.Column(sqlalchemy.Integer)
        visionWardsBoughtInGame = sqlalchemy.Column(sqlalchemy.Integer)
        wardsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        wardsPlaced = sqlalchemy.Column(sqlalchemy.Integer)
        winner = sqlalchemy.Column(sqlalchemy.Boolean)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipant._id", ondelete="CASCADE"))


def _sa_bind_participant_timeline():
    global ParticipantTimeline

    @cassiopeia.type.core.common.inheritdocs
    class ParticipantTimeline(ParticipantTimeline, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchParticipantTimeline"
        ancientGolemAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='ancientGolemAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        ancientGolemKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='ancientGolemKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        assistedLaneDeathsPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='assistedLaneDeathsPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        assistedLaneKillsPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='assistedLaneKillsPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        baronAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='baronAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        baronKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='baronKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        creepsPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='creepsPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        csDiffPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='csDiffPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        damageTakenDiffPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='damageTakenDiffPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        damageTakenPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='damageTakenPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        dragonAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='dragonAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        dragonKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='dragonKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        elderLizardAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='elderLizardAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        elderLizardKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='elderLizardKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        goldPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='goldPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        inhibitorAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='inhibitorAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        inhibitorKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='inhibitorKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        lane = sqlalchemy.Column(sqlalchemy.String(30))
        role = sqlalchemy.Column(sqlalchemy.String(30))
        towerAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='towerAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        towerKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='towerKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        towerKillsPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='towerKillsPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        vilemawAssistsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='vilemawAssistsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        vilemawKillsPerMinCounts = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='vilemawKillsPerMinCounts')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        wardsPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='wardsPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        xpDiffPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='xpDiffPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        xpPerMinDeltas = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantTimelineData", primaryjoin="and_(cassiopeia.type.dto.match.ParticipantTimeline._id==cassiopeia.type.dto.match.ParticipantTimelineData._timeline_id, cassiopeia.type.dto.match.ParticipantTimelineData._type=='xpPerMinDeltas')", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipant._id", ondelete="CASCADE"))


def _sa_bind_rune():
    global Rune

    @cassiopeia.type.core.common.inheritdocs
    class Rune(Rune, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchRune"
        rank = sqlalchemy.Column(sqlalchemy.Integer)
        runeId = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipant._id", ondelete="CASCADE"))


def _sa_bind_player():
    global Player

    @cassiopeia.type.core.common.inheritdocs
    class Player(Player, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchPlayer"
        matchHistoryUri = sqlalchemy.Column(sqlalchemy.String(50))
        profileIcon = sqlalchemy.Column(sqlalchemy.Integer)
        summonerId = sqlalchemy.Column(sqlalchemy.Integer)
        summonerName = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipantIdentity._id", ondelete="CASCADE"))


def _sa_bind_banned_champion():
    global BannedChampion

    @cassiopeia.type.core.common.inheritdocs
    class BannedChampion(BannedChampion, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchBannedChampion"
        championId = sqlalchemy.Column(sqlalchemy.Integer)
        pickTurn = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _team_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchTeam._id", ondelete="CASCADE"))


def _sa_bind_frame():
    global Frame

    @cassiopeia.type.core.common.inheritdocs
    class Frame(Frame, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchFrame"
        events = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Event", cascade="all, delete-orphan", passive_deletes=True)
        participantFrames = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.ParticipantFrame", collection_class=sqlalchemy.orm.collections.mapped_collection(lambda p: str(p.participantId)), cascade="all, delete-orphan", passive_deletes=True)  # OR I HAVE NO IDEA
        timestamp = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _timeline_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchTimeline._id", ondelete="CASCADE"))


def _sa_bind_participant_timeline_data():
    global ParticipantTimelineData

    @cassiopeia.type.core.common.inheritdocs
    class ParticipantTimelineData(ParticipantTimelineData, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchParticipantTimelineData"
        tenToTwenty = sqlalchemy.Column(sqlalchemy.Float)
        thirtyToEnd = sqlalchemy.Column(sqlalchemy.Float)
        twentyToThirty = sqlalchemy.Column(sqlalchemy.Float)
        zeroToTen = sqlalchemy.Column(sqlalchemy.Float)
        _type = sqlalchemy.Column(sqlalchemy.String(50))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _timeline_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipantTimeline._id", ondelete="CASCADE"))


def _sa_bind_event():
    global Event

    @cassiopeia.type.core.common.inheritdocs
    class Event(Event, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchEvent"
        ascendedType = sqlalchemy.Column(sqlalchemy.String(30))
        assistingParticipantIds = sqlalchemy.Column(cassiopeia.type.dto.common.JSONEncoded)
        buildingType = sqlalchemy.Column(sqlalchemy.String(30))
        creatorId = sqlalchemy.Column(sqlalchemy.Integer)
        eventType = sqlalchemy.Column(sqlalchemy.String(30))
        itemAfter = sqlalchemy.Column(sqlalchemy.Integer)
        itemBefore = sqlalchemy.Column(sqlalchemy.Integer)
        itemId = sqlalchemy.Column(sqlalchemy.Integer)
        killerId = sqlalchemy.Column(sqlalchemy.Integer)
        laneType = sqlalchemy.Column(sqlalchemy.String(30))
        levelUpType = sqlalchemy.Column(sqlalchemy.String(30))
        monsterType = sqlalchemy.Column(sqlalchemy.String(30))
        participantId = sqlalchemy.Column(sqlalchemy.Integer)
        pointCaptured = sqlalchemy.Column(sqlalchemy.String(30))
        position = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Position", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        skillSlot = sqlalchemy.Column(sqlalchemy.Integer)
        teamId = sqlalchemy.Column(sqlalchemy.Integer)
        timestamp = sqlalchemy.Column(sqlalchemy.Integer)
        towerType = sqlalchemy.Column(sqlalchemy.String(30))
        victimId = sqlalchemy.Column(sqlalchemy.Integer)
        wardType = sqlalchemy.Column(sqlalchemy.String(30))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _frame_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchFrame._id", ondelete="CASCADE"))


def _sa_bind_participant_frame():
    global ParticipantFrame

    @cassiopeia.type.core.common.inheritdocs
    class ParticipantFrame(ParticipantFrame, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchParticipantFrame"
        currentGold = sqlalchemy.Column(sqlalchemy.Integer)
        dominionScore = sqlalchemy.Column(sqlalchemy.Integer)
        jungleMinionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        level = sqlalchemy.Column(sqlalchemy.Integer)
        minionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        participantId = sqlalchemy.Column(sqlalchemy.Integer)
        position = sqlalchemy.orm.relationship("cassiopeia.type.dto.match.Position", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        teamScore = sqlalchemy.Column(sqlalchemy.Integer)
        totalGold = sqlalchemy.Column(sqlalchemy.Integer)
        xp = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _frame_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchFrame._id", ondelete="CASCADE"))


def _sa_bind_position():
    global Position

    @cassiopeia.type.core.common.inheritdocs
    class Position(Position, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "MatchPosition"
        x = sqlalchemy.Column(sqlalchemy.Integer)
        y = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _event_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchEvent._id", ondelete="CASCADE"))
        _frame_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("MatchParticipantFrame._id", ondelete="CASCADE"))


def _sa_bind_all():
    _sa_bind_match_detail()
    _sa_bind_participant()
    _sa_bind_participant_identity()
    _sa_bind_team()
    _sa_bind_timeline()
    _sa_bind_mastery()
    _sa_bind_participant_stats()
    _sa_bind_participant_timeline()
    _sa_bind_rune()
    _sa_bind_player()
    _sa_bind_banned_champion()
    _sa_bind_frame()
    _sa_bind_participant_timeline_data()
    _sa_bind_event()
    _sa_bind_participant_frame()
    _sa_bind_position()



import datetime

import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.game

try:
    from future.builtins.misc import super
except ImportError:
    pass


@cassiopeia.type.core.common.inheritdocs
class Stats(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.game.RawStats

    def __str__(self):
        return "Stats"

    @property
    def kda(self):
        """
        Returns:
            float: the participant's kda
        """
        return (self.kills + self.assists) / (self.deaths if self.deaths else 1)

    @property
    def assists(self):
        """
        Returns:
            int: the total number of assists this participant had
        """
        return self.data.assists

    @property
    def inhibitor_kills(self):
        """
        Returns:
            int: the total number of inhibitors this participant killed
        """
        return self.data.barracksKilled

    @property
    def kills(self):
        """
        Returns:
            int: the total number of kills this participant had
        """
        return self.data.championsKilled

    @property
    def combat_score(self):
        """
        Returns:
            int: dominion only. the part of the participant's score that came from combat-related activities
        """
        return self.data.combatPlayerScore

    @property
    def consumables_bought(self):
        """
        Returns:
            list<Item>: the consumables that the participant bought (careful, they might have just sold them back or hit undo?)
        """
        return self.data.consumablesPurchased

    @property
    def damage_dealt_player(self):
        """
        Returns:
            int: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.damageDealtPlayer

    @property
    def double_kills(self):
        """
        Returns:
            int: the number of double kills this participant had
        """
        return self.data.doubleKills

    @property
    def first_blood(self):
        """
        Returns:
            bool: whether this participant got first blood
        """
        return self.data.firstBlood

    @property
    def gold(self):
        """
        Returns:
            int: the participant's current gold
        """
        return self.data.gold

    @property
    def gold_earned(self):
        """
        Returns:
            int: the participant's total gold
        """
        return self.data.goldEarned

    @property
    def gold_spent(self):
        """
        Returns:
            int: the participant's spent gold
        """
        return self.data.goldSpent

    @property
    def item0(self):
        """
        Returns:
            Item: the participant's first item
        """
        return cassiopeia.riotapi.get_item(self.data.item0) if self.data.item0 else None

    @property
    def item1(self):
        """
        Returns:
            Item: the participant's second item
        """
        return cassiopeia.riotapi.get_item(self.data.item1) if self.data.item1 else None

    @property
    def item2(self):
        """
        Returns:
            Item: the participant's third item
        """
        return cassiopeia.riotapi.get_item(self.data.item2) if self.data.item2 else None

    @property
    def item3(self):
        """
        Returns:
            Item: the participant's fourth item
        """
        return cassiopeia.riotapi.get_item(self.data.item3) if self.data.item3 else None

    @property
    def item4(self):
        """
        Returns:
            Item: the participant's fifth item
        """
        return cassiopeia.riotapi.get_item(self.data.item4) if self.data.item4 else None

    @property
    def item5(self):
        """
        Returns:
            Item: the participant's sixth item
        """
        return cassiopeia.riotapi.get_item(self.data.item5) if self.data.item5 else None

    @property
    def item6(self):
        """
        Returns:
            Item: the participant's seventh item (i.e. their ward)
        """
        return cassiopeia.riotapi.get_item(self.data.item6) if self.data.item6 else None

    @property
    def items(self):
        """
        Returns:
            list<Item>: the participant's items
        """
        return [self.item0, self.item1, self.item2, self.item3, self.item4, self.item5, self.item6]

    @property
    def killing_sprees(self):
        """
        Returns:
            int: the number of killing sprees this participant had
        """
        return self.data.killingSprees

    @property
    def largest_critical_strike(self):
        """
        Returns:
            int: the largest critical strike this participant had
        """
        return self.data.largestCriticalStrike

    @property
    def largest_killing_spree(self):
        """
        Returns:
            int: the larges killing spree this participant had
        """
        return self.data.largestKillingSpree

    @property
    def largest_multi_kill(self):
        """
        Returns:
            int: the largest multikill this participant had
        """
        return self.data.largestMultiKill

    @property
    def tier_3_items_bought(self):
        """
        Returns:
            int: the number of tier 3 items built
        """
        return self.data.legendaryItemsCreated

    @property
    def level(self):
        """
        Returns:
            int: the participant's champion level
        """
        return self.data.level

    @property
    def magic_damage_dealt(self):
        """
        Returns:
            int: the total magic damage this participant dealt
        """
        return self.data.magicDamageDealtPlayer

    @property
    def magic_damage_dealt_to_champions(self):
        """
        Returns:
            int: the total magic damage this participant dealt to champions
        """
        return self.data.magicDamageDealtToChampions

    @property
    def magic_damage_taken(self):
        """
        Returns:
            int: the total magic damage this participant received
        """
        return self.data.magicDamageTaken

    @property
    def minion_denies(self):
        """
        Returns:
            int: the number of minions this participant denied to the enemy. let us know if you figure out what this actually is
        """
        return self.data.minionsDenied

    @property
    def minion_kills(self):
        """
        Returns:
            int: the number of minions this participant killed
        """
        return self.data.minionsKilled

    @property
    def monster_kills(self):
        """
        Returns:
            int: the number of neutral minions this participant killed
        """
        return self.data.neutralMinionsKilled

    @property
    def enemy_monster_kills(self):
        """
        Returns:
            int: the number of neutral enemy minions this participant killed
        """
        return self.data.neutralMinionsKilledEnemyJungle

    @property
    def ally_monster_kills(self):
        """
        Returns:
            int: the number of neutral ally minions this participant killed
        """
        return self.data.neutralMinionsKilledYourJungle

    @property
    def nexus_killed(self):
        """
        Returns:
            int: the number of nexuses this participant killed
        """
        return self.data.nexusKilled

    @property
    def node_captured(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant captured
        """
        return self.data.nodeCapture

    @property
    def node_capture_assists(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant assisted in capturing
        """
        return self.data.nodeCaptureAssist

    @property
    def node_neutralizations(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant neutralized
        """
        return self.data.nodeNeutralize

    @property
    def node_neutralization_assists(self):
        """
        Returns:
            int: dominion only. the number of nodes this participant assisted in neutralizing
        """
        return self.data.nodeNeutralizeAssist

    @property
    def deaths(self):
        """
        Returns:
            int: the number of deaths this participant had
        """
        return self.data.numDeaths

    @property
    def items_bought(self):
        """
        Returns:
            int: the number of items this participant bought
        """
        return self.data.numItemsBought

    @property
    def objective_score(self):
        """
        Returns:
            int: dominion only. the part of the participant's score that came from objective-related activities
        """
        return self.data.objectivePlayerScore

    @property
    def penta_kills(self):
        """
        Returns:
            int: the number of penta kills this participant had
        """
        return self.data.pentaKills

    @property
    def physical_damage_dealt(self):
        """
        Returns:
            int: the total physical damage this participant dealt
        """
        return self.data.physicalDamageDealtPlayer

    @property
    def physical_damage_dealt_to_champions(self):
        """
        Returns:
            int: the total physical damage this participant dealt to champions
        """
        return self.data.physicalDamageDealtToChampions

    @property
    def physical_damage_taken(self):
        """
        Returns:
            int: the total physical damage this participant received
        """
        return self.data.physicalDamageTaken

    @property
    def lane(self):
        """
        Returns:
            Lane: the lane this participant was in
        """
        return cassiopeia.type.core.common.Lane.for_id(self.data.playerPosition) if self.data.playerPosition else None

    @property
    def role(self):
        """
        Returns:
            Role: the role of this particiant
        """
        return cassiopeia.type.core.common.Role.for_id(self.data.playerRole) if self.data.playerRole else None

    @property
    def quadra_kills(self):
        """
        Returns:
            int: the number of quadra kills this participant had
        """
        return self.data.quadraKills

    @property
    def sight_wards_bought(self):
        """
        Returns:
            int: the number of sight wards this participant bought
        """
        return self.data.sightWardsBought

    @property
    def q_casts(self):
        """
        Returns:
            int: the number of times this participant cast his Q
        """
        return self.data.spell1Cast

    @property
    def w_casts(self):
        """
        Returns:
            int: the number of tiems this participant cast his W
        """
        return self.data.spell2Cast

    @property
    def e_casts(self):
        """
        Returns:
            int: the number of times this participant cast his E
        """
        return self.data.spell3Cast

    @property
    def r_casts(self):
        """
        Returns:
            int: the number of times this participant cast his R
        """
        return self.data.spell4Cast

    @property
    def d_casts(self):
        """
        Returns:
            int: the number of times this participant cast his D summoner spell
        """
        return self.data.summonSpell1Cast

    @property
    def f_casts(self):
        """
        Returns:
            int: the number of times this participant cast his F summoner spell
        """
        return self.data.summonSpell2Cast

    @property
    def elite_monsters_kills(self):
        """
        Returns:
            int: the number of elite monsters this participant killed
        """
        return self.data.superMonsterKilled

    @property
    def side(self):
        """
        Returns:
            Side: the side the participant was on
        """
        return cassiopeia.type.core.common.Side(self.data.team) if self.data.team else None

    @property
    def objectives(self):
        """
        Returns:
            int: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.teamObjective

    @property
    def time_played(self):
        """
        Returns:
            int: the amount of time this participant played
        """
        return self.data.timePlayed

    @property
    def damage_dealt(self):
        """
        Returns:
            int: the total damage this participant dealt
        """
        return self.data.totalDamageDealt

    @property
    def damage_dealt_to_champions(self):
        """
        Returns:
            int: the total damage this participant dealt to champions
        """
        return self.data.totalDamageDealtToChampions

    @property
    def damage_taken(self):
        """
        Returns:
            int: the total damage this participant received
        """
        return self.data.totalDamageTaken

    @property
    def healing_done(self):
        """
        Returns:
            int: the amount of healing this participant did
        """
        return self.data.totalHeal

    @property
    def score(self):
        """
        Returns:
            int: the score for this participant
        """
        return self.data.totalPlayerScore

    @property
    def score_rank(self):
        """
        Returns:
            int: if game was a dominion game, team rank of the player's total score (e.g., 1-5)
        """
        return self.data.totalScoreRank

    @property
    def crowd_control_dealt(self):
        """
        Returns:
            int: the total amount of crowd control this participant dealt (in seconds)
        """
        return self.data.totalTimeCrowdControlDealt

    @property
    def units_healed(self):
        """
        Returns:
            int: the number of units this participant healed
        """
        return self.data.totalUnitsHealed

    @property
    def triple_kills(self):
        """
        Returns:
            int: the number of triple kills this participant had
        """
        return self.data.tripleKills

    @property
    def true_damage_dealt(self):
        """
        Returns:
            int: the total true damage this participant dealth
        """
        return self.data.trueDamageDealtPlayer

    @property
    def true_damage_dealt_to_champions(self):
        """
        Returns:
            int: the total damage this participant dealt to champions
        """
        return self.data.trueDamageDealtToChampions

    @property
    def true_damage_taken(self):
        """
        Returns:
            int: the total true damage this participant received
        """
        return self.data.trueDamageTaken

    @property
    def turret_kills(self):
        """
        Returns:
            int: the number of turret kills this participant had
        """
        return self.data.turretsKilled

    @property
    def unreal_kills(self):
        """
        Returns:
            int: the number of unreal kills this participant had
        """
        return self.data.unrealKills

    @property
    def victory_points(self):
        """
        Returns:
            int: the number of victory points this participant gained from winning or losing this game
        """
        return self.data.victoryPointTotal

    @property
    def vision_wards_bought(self):
        """
        Returns:
            int: the number of vision wards sprees this participant bought
        """
        return self.data.visionWardsBought

    @property
    def ward_kills(self):
        """
        Returns:
            int: the number of wards sprees this participant killed
        """
        return self.data.wardKilled

    @property
    def wards_placed(self):
        """
        Returns:
            int: the number of wards this participant placed
        """
        return self.data.wardPlaced

    @property
    def win(self):
        """
        Returns:
            bool: whether the participant won the game or not
        """
        return self.data.win


@cassiopeia.type.core.common.inheritdocs
class Participant(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.game.Player

    def __str__(self):
        return "{player} ({champ})".format(player=self.summoner, champ=self.champion)

    @property
    def champion(self):
        """
        Returns:
            Champion: the champion for this participant
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.championId) if self.data.championId else None

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the summoner for this participant
        """
        return cassiopeia.riotapi.get_summoner_by_id(self.data.summonerId) if self.data.summonerId else None

    @property
    def side(self):
        """
        Returns:
            Side: the side the participant was on
        """
        return cassiopeia.type.core.common.Side(self.data.teamId) if self.data.teamId else None


@cassiopeia.type.core.common.inheritdocs
class Game(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.game.Game

    def __init__(self, data, summoner_id):
        super().__init__(data)
        self.__summoner_id = summoner_id

    def __str__(self):
        return "Game #{id}".format(id=self.id)

    def __iter__(self):
        return iter(self.participants)

    def __len__(self):
        return len(self.participants)

    def __getitem__(self, index):
        return self.participants[index]

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the summoner for this participant
        """
        return cassiopeia.riotapi.get_summoner_by_id(self.__summoner_id) if self.__summoner_id else None

    @property
    def champion(self):
        """
        Returns:
            Champion: the champion for this participant
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.championId) if self.data.championId else None

    @cassiopeia.type.core.common.lazyproperty
    def creation(self):
        """
        Returns:
            datetime: the time when this game was created
        """
        return datetime.datetime.utcfromtimestamp(self.data.createDate / 1000) if self.data.createDate else None

    @cassiopeia.type.core.common.lazyproperty
    def participants(self):
        """
        Returns:
            list<Participant>: the participants in this game
        """
        parts = [Participant(player) for player in self.data.fellowPlayers]
        parts.append(Participant(cassiopeia.type.dto.game.Player({
            "championId": self.data.championId,
            "summonerId": self.__summoner_id,
            "teamId": self.data.teamId
        })))
        return parts

    @property
    def id(self):
        """
        Returns:
            int: the match ID
        """
        return self.data.gameId

    @property
    def mode(self):
        """
        Returns:
            GameMode: the game mode
        """
        return cassiopeia.type.core.common.GameMode(self.data.gameMode) if self.data.gameMode else None

    @property
    def type(self):
        """
        Returns:
            GameType: the game type
        """
        return cassiopeia.type.core.common.GameType(self.data.gameType) if self.data.gameType else None

    @property
    def invalid(self):
        """
        Returns:
            bool: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.invalid

    @property
    def ip(self):
        """
        Returns:
            int: the amount of IP the participant gained for this game (the one that this game was pulled using)
        """
        return self.data.ipEarned

    @property
    def level(self):
        """
        Returns:
            int: the participant's champion level
        """
        return self.data.level

    @property
    def map(self):
        """
        Returns:
            Map: the map this game was played on
        """
        return cassiopeia.type.core.common.Map(self.data.mapId) if self.data.mapId else None

    @property
    def summoner_spell_d(self):
        """
        Returns:
            SummonerSpell: the particpant's first summoner spell (the one that this game was pulled using)
        """
        return cassiopeia.riotapi.get_summoner_spell(self.data.spell1) if self.data.spell1 else None

    @property
    def summoner_spell_f(self):
        """
        Returns:
            SummonerSpell: the participant's second summoner spell (the one that this game was pulled using)
        """
        return cassiopeia.riotapi.get_summoner_spell(self.data.spell2) if self.data.spell2 else None

    @cassiopeia.type.core.common.lazyproperty
    def stats(self):
        """
        Returns:
            Stats: the participant's stats (the one that this game was pulled using)
        """
        return Stats(self.data.stats) if self.data.stats else None

    @property
    def sub_type(self):
        """
        Returns:
            SubType: the game's sub-type
        """
        return cassiopeia.type.core.common.SubType(self.data.subType) if self.data.subType else None

    @property
    def side(self):
        """
        Returns:
            Side: the side the participant was on
        """
        return cassiopeia.type.core.common.Side(self.data.teamId) if self.data.teamId else None


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    Stats.dto_type = cassiopeia.type.dto.game.RawStats
    Participant.dto_type = cassiopeia.type.dto.game.Player
    Game.dto_type = cassiopeia.type.dto.game.Game



import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.league


@cassiopeia.type.core.common.inheritdocs
class Series(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.league.MiniSeries

    def __str__(self):
        return self.progress

    @property
    def losses(self):
        """
        Returns:
            int: number of current losses in the mini series
        """
        return self.data.losses

    @property
    def progress(self):
        """
        Returns:
            str: string showing the current, sequential mini series progress where 'W' represents a win, 'L' epresents a loss, and 'N' represents a game that hasn't been played yet
        """
        return self.data.progress

    @property
    def wins_required(self):
        """
        Returns:
            int: number of wins required for promotion
        """
        return self.data.target

    @property
    def wins(self):
        """
        Returns:
            int: number of current wins in the mini series
        """
        return self.data.wins


@cassiopeia.type.core.common.inheritdocs
class Entry(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.league.LeagueEntry

    def __str__(self):
        return "{summoner} ({lp} LP)".format(summoner=self.summoner_name, lp=self.league_points)

    @property
    def division(self):
        """
        Returns:
            Division: the league division of the participant
        """
        return cassiopeia.type.core.common.Division(self.data.division) if self.data.division else None

    @property
    def fresh_blood(self):
        """
        Returns:
            bool: specifies if the participant is fresh blood (ie if they have just joined the league)
        """
        return self.data.isFreshBlood

    @property
    def hot_streak(self):
        """
        Returns:
            bool: specifies if the participant is on a hot streak
        """
        return self.data.isHotStreak

    @property
    def inactive(self):
        """
        Returns:
            bool: specifies if the participant is inactive
        """
        return self.data.isInactive

    @property
    def veteran(self):
        """
        Returns:
            bool: specifies if the participant is a veteran (ie they have been in this league for a long time)
        """
        return self.data.isVeteran

    @property
    def league_points(self):
        """
        Returns:
            int: the league points of the participant
        """
        return self.data.leaguePoints

    @property
    def losses(self):
        """
        Returns:
            int: number of current losses for the participant
        """
        return self.data.losses

    @cassiopeia.type.core.common.lazyproperty
    def series(self):
        """
        Returns:
            Series: series data for the participant. Only present if the participant is currently in a mini series
        """
        return Series(self.data.miniSeries) if self.data.miniSeries else None

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the summoner represented by this entry. None if this entry is for a team
        """
        if not self.data.playerOrTeamId:
            return None

        try:
            id_ = int(self.data.playerOrTeamId)
            return cassiopeia.riotapi.get_summoner_by_id(id_)
        except ValueError:
            return None

    @property
    def team(self):
        """
        Returns:
            Team: the team represented by this entry. None if this entry is for a summoner
        """
        if not self.data.playerOrTeamId:
            return None

        try:
            int(self.data.playerOrTeamId)
            return None
        except ValueError:
            return cassiopeia.riotapi.get_team_by_id(self.data.playerOrTeamId)

    @property
    def summoner_name(self):
        """
        Returns:
            str: the name of the summoner represented by this entry. An empty string if this entry is for a team
        """
        if not self.data.playerOrTeamId:
            return ""

        try:
            int(self.data.playerOrTeamId)
            return self.data.playerOrTeamName
        except ValueError:
            return ""

    @property
    def team_name(self):
        """
        Returns:
            str: the name of the team represented by this entry. An empty string if this entry is for a summoner
        """
        try:
            int(self.data.playerOrTeamId)
            return ""
        except ValueError:
            return self.data.playerOrTeamName

    @property
    def wins(self):
        """
        Returns:
            int: the number of wins for the participant
        """
        return self.data.wins


@cassiopeia.type.core.common.inheritdocs
class League(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.league.League

    def __str__(self):
        return "{name} ({tier})".format(name=self.name, tier=self.tier)

    def __iter__(self):
        return iter(self.entries)

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, index):
        return self.entries[index]

    @cassiopeia.type.core.common.lazyproperty
    def entries(self):
        """
        Returns:
            list<Entry>: a list of the requested league entries, sorted by LP
        """
        return sorted([Entry(entry) for entry in self.data.entries], key=lambda entry: entry.league_points, reverse=True)

    @property
    def name(self):
        """
        Returns:
            str: the name of the league
        """
        return self.data.name

    @cassiopeia.type.core.common.lazyproperty
    def participant_entry(self):
        """
        Returns:
            Entry: the entry for the relevant team or summoner that is a member of this league. Only present when full league is requested so that participant's entry can be identified. None when individual entry is requested
        """
        for entry in self.entries:
            if entry.data.playerOrTeamId == self.data.participantId:
                return entry
        return None

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the relevant summoner that is a member of this league. Only present when full league is requested so that participant's entry can be identified. None when individual entry is requested or the participant is a team.
        """
        if not self.data.participantId:
            return None

        try:
            id_ = int(self.data.participantId)
            return cassiopeia.riotapi.get_summoner_by_id(id_)
        except ValueError:
            return None

    @property
    def team(self):
        """
        Returns:
            Team: the relevant team that is a member of this league. Only present when full league is requested so that participant's entry can be identified. None when individual entry is requested or the participant is a summoner.
        """
        if not self.data.participantId:
            return None

        try:
            int(self.data.participantId)
            return None
        except ValueError:
            return cassiopeia.riotapi.get_team_by_id(self.data.participantId)

    @property
    def queue(self):
        """
        Returns:
            Queue: the league's queue type
        """
        return cassiopeia.type.core.common.Queue(self.data.queue) if self.data.queue else None

    @property
    def tier(self):
        """
        Returns:
            Tier: the league's tier
        """
        return cassiopeia.type.core.common.Tier(self.data.tier) if self.data.tier else None


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    Series.dto_type = cassiopeia.type.dto.league.MiniSeries
    Entry.dto_type = cassiopeia.type.dto.league.LeagueEntry
    League.dto_type = cassiopeia.type.dto.league.League



import threading


class SingleRateLimiter(object):
    """
    Handles a single rate limit, ensuring that calls don't exceed it
    """

    def __init__(self, calls_per_epoch, seconds_per_epoch):
        """
        Handles a single rate limit, ensuring that calls don't exceed it
        """
        self.seconds_per_epoch = seconds_per_epoch
        self.semaphore = threading.Semaphore(calls_per_epoch)
        self.lock = threading.Lock()
        self.current = 0
        self.limit = calls_per_epoch
        self.resetter = None
        self._total_calls = 0
        self._successful_calls = 0

    def call(self, method=None, *args):
        """
        Args:
            calls_per_epoch (int): the number of calls allowed in each epoch
            seconds_per_epoch (int): the number of seconds per epoch
        """
        # Block until a call opens up
        self.semaphore.acquire()

        # Current tracks how many calls are currently inside the method body
        with self.lock:
            self.current += 1

        successful_call = True
        try:
            return method(*args) if method else None
        except:
            successful_call = False
            raise
        finally:
            # If we don't have an epoch timer running, start one. Also count that we've completed a call.
            with self.lock:
                if not self.resetter:
                    self.resetter = threading.Timer(self.seconds_per_epoch, self._reset)
                    self.resetter.daemon = True
                    self.resetter.start()

                self.current -= 1
                self._total_calls += 1
                if successful_call:
                    self._successful_calls += 1

    def _drain(self):
        """
        Calls a function when the rate limit allows (first come first serve)

        Args:
            method (function): the function which will be called when the rate limit allows
            *args (any...): the arguments to be passed to the functions when it is called

        Returns:
            any: the result of the function once it has been called
        """
        while self.semaphore.acquire(False):
            pass

    def _reset(self):
        """
        Calls a function when the rate limit allows (first come first serve)

        Args:
            method (function): the function which will be called when the rate limit allows
            *args (any...): the arguments to be passed to the functions when it is called

        Returns:
            any: the result of the function once it has been called
        """
        self.lock.acquire()

        self._drain()
        for _ in range(self.limit - self.current):
            self.semaphore.release()
        self.resetter = None

        self.lock.release()

    def wait(self):
        """
        Calls a function when the rate limit allows (first come first serve)

        Args:
            method (function): the function which will be called when the rate limit allows
            *args (any...): the arguments to be passed to the functions when it is called

        Returns:
            any: the result of the function once it has been called
        """
        self.semaphore.acquire()
        self.semaphore.release()

    def reset_in(self, seconds):
        """
        Calls a function when the rate limit allows (first come first serve)

        Args:
            method (function): the function which will be called when the rate limit allows
            *args (any...): the arguments to be passed to the functions when it is called

        Returns:
            any: the result of the function once it has been called
        """
        with self.lock:

            if self.resetter:
                self.resetter.cancel()

            self._drain()
            self.resetter = threading.Timer(seconds, self._reset)
            self.resetter.daemon = True
            self.resetter.start()

    def _decrease_successful_calls(self):
        with self.lock:
            self._successful_calls -= 1

    @property
    def calls(self):
        """
        Drains all remaining calls
        """
        with self.lock:
            return (self._successful_calls, self._total_calls)


class MultiRateLimiter(object):
    """
    Resets the rate limit
    """

    def __init__(self, *limits):
        """
        Resets the rate limit
        """
        self.limits = []

        for limit in limits:
            self.limits.append(SingleRateLimiter(limit[0], limit[1]))

    def call(self, method=None, *args):  # TODO
        """
        Waits until a call becomes available
        """
        self.wait()

        successful_call = True
        try:
            return method(*args) if method else None
        except:
            successful_call = False
            raise
        finally:
            for limit in self.limits:
                # call with method=None always updates the successful calls
                limit.call()
                if not successful_call:
                    limit._decrease_successful_calls()

    def wait(self):
        """
        Resets the rate limiter after waiting

        Args:
            seconds (int): the number of seconds to wait before resetting
        """
        for limit in self.limits:
            limit.wait()

    def reset_in(self, seconds):
        """
        Resets the rate limiter after waiting

        Args:
            seconds (int): the number of seconds to wait before resetting
        """
        for limit in self.limits:
            limit.reset_in(seconds)

    @property
    def calls(self):
        """
        Returns the number of successful calls (no exceptions in the call) and total calls served by this limiter

        Returns:
            return: tuple   A (successful calls, total calls) tuple
        """
        try:
            return self.limits[0].calls
        except IndexError:
            # Disable the counting if no limits are enforced
            return 0



import cassiopeia.riotapi
import cassiopeia.dto.summonerapi
import cassiopeia.core.requests
import cassiopeia.type.core.common
import cassiopeia.type.core.summoner


def __get_mastery_pages_by_id(ids):
    pages = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.summonerapi.get_summoner_masteries, 40, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        mastery_ids = set()
        for page in pages.values():
            mastery_ids |= page.mastery_ids
        cassiopeia.riotapi.get_masteries() if mastery_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.summoner.MasteryPage(page) for page in pages[str(ids)].pages]
    else:
        return [[cassiopeia.type.core.summoner.MasteryPage(page) for page in pages[str(id_)].pages] for id_ in ids]


def get_mastery_pages(summoners):
    """
    Get the mastery pages for (a) summoner(s).

    Args:
        ids (Summoner | list<Summoner>): the summoner(s) to get mastery pages for

    Returns:
        list<MasteryPage> | list<list<MasteryPage>>: the requested summoner(s)' mastery pages
    """
    if isinstance(summoners, list):
        return __get_mastery_pages_by_id([summoner.id for summoner in summoners])
    else:
        return __get_mastery_pages_by_id(summoners.id)


def __get_rune_pages_by_id(ids):
    pages = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.summonerapi.get_summoner_runes, 40, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        rune_ids = set()
        for page in pages.values():
            rune_ids |= page.rune_ids
        cassiopeia.riotapi.get_runes() if rune_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.summoner.RunePage(page) for page in pages[str(ids)].pages]
    else:
        return [[cassiopeia.type.core.summoner.RunePage(page) for page in pages[str(id_)].pages] for id_ in ids]


def get_rune_pages(summoners):
    """
    Get the rune pages for (a) summoner(s).

    Args:
        ids (Summoner | list<Summoner>): the summoner(s) to get rune pages for

    Returns:
        list<RunePage> | list<list<RunePage>>: the requested summoner(s)' rune pages
    """
    if isinstance(summoners, list):
        return __get_rune_pages_by_id([summoner.id for summoner in summoners])
    else:
        return __get_rune_pages_by_id(summoners.id)


def get_summoner_by_id(id_):
    """
    Gets a summoner by ID

    Args:
        id_ (int): the ID of the summoner

    Returns:
        Summoner: the summoner
    """
    summoner = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.summoner.Summoner, id_, "id")
    if summoner:
        return summoner

    try:
        summoner = cassiopeia.dto.summonerapi.get_summoners_by_id(id_)[str(id_)]
        summoner = cassiopeia.type.core.summoner.Summoner(summoner)
    except KeyError:
        return None

    cassiopeia.core.requests.data_store.store(summoner, id_)
    cassiopeia.core.requests.data_store.store(summoner, summoner.name)
    return summoner


def get_summoner_by_name(name):
    """
    Gets a summoner by name

    Args:
        name (str): the name of the summoner

    Returns:
        Summoner: the summoner
    """
    summoner = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.summoner.Summoner, name, "name")
    if summoner:
        return summoner

    try:
        summoner = cassiopeia.dto.summonerapi.get_summoners_by_name(name)[__standardize(name)]
        summoner = cassiopeia.type.core.summoner.Summoner(summoner)
    except KeyError:
        return None

    cassiopeia.core.requests.data_store.store(summoner, name)
    cassiopeia.core.requests.data_store.store(summoner, summoner.id)
    return summoner


def get_summoners_by_id(ids):
    """
    Gets a bunch of summoners by ID

    Args:
        ids (list<int>): the IDs of the summoners

    Returns:
        list<Summoner>: the summoners
    """
    summoners = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.summoner.Summoner, ids, "id")

    # Find which summoners weren't cached
    missing = []
    loc = []
    for i in range(len(ids)):
        if not summoners[i]:
            missing.append(ids[i])
            loc.append(i)

    if not missing:
        return summoners

    # Make requests to get them
    new = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.summonerapi.get_summoners_by_id, 40, missing)
    to_store = []
    for i in range(len(missing)):
        try:
            summoner = cassiopeia.type.core.summoner.Summoner(new[str(missing[i])])
            to_store.append(summoner)
        except KeyError:
            summoner = None
        summoners[loc[i]] = summoner

    cassiopeia.core.requests.data_store.store(to_store, [summoner.id for summoner in to_store])
    cassiopeia.core.requests.data_store.store(to_store, [summoner.name for summoner in to_store])
    return summoners


def get_summoners_by_name(names):
    """
    Gets a bunch of summoners by name

    Args:
        names (list<str>): the names of the summoners

    Returns:
        list<Summoner>: the summoners
    """
    summoners = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.summoner.Summoner, names, "name")

    # Find which summoners weren't cached
    missing = []
    loc = []
    for i in range(len(names)):
        if not summoners[i]:
            missing.append(names[i])
            loc.append(i)

    if not missing:
        return summoners

    # Make requests to get them
    new = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.summonerapi.get_summoners_by_name, 40, missing)
    to_store = []
    for i in range(len(missing)):
        try:
            summoner = cassiopeia.type.core.summoner.Summoner(new[__standardize(missing[i])])
            to_store.append(summoner)
        except KeyError:
            summoner = None
        summoners[loc[i]] = summoner

    cassiopeia.core.requests.data_store.store(to_store, [summoner.id for summoner in to_store])
    cassiopeia.core.requests.data_store.store(to_store, [summoner.name for summoner in to_store])
    return summoners


def get_summoner_name(id_):
    """
    Gets the name of a summoner by ID

    Args:
        id_ (id): the summoner's ID

    Returns:
        str: the summoner's name
    """
    summoner = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.summoner.Summoner, id_, "id")
    if summoner:
        return summoner.name

    return cassiopeia.dto.summonerapi.get_summoner_names(id_)[str(id_)]


def get_summoner_names(ids):
    """
    Gets the names of a bunch of summoners by ID

    Args:
        ids (list<id>): the summoners' IDs

    Returns:
        list<str>: the summoners' names
    """
    summoners = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.summoner.Summoner, ids, "id")
    summoners = [summoner.name if summoner else "" for summoner in summoners]

    # Find which summoners weren't cached
    missing = []
    loc = []
    for i in range(len(ids)):
        if not summoners[i]:
            missing.append(ids[i])
            loc.append(i)

    if not missing:
        return summoners

    # Make requests to get them
    new = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.summonerapi.get_summoner_names, 40, missing)
    for i in range(len(missing)):
        summoners[loc[i]] = new[str(missing[i])]

    return summoners


def __standardize(name):
    return name.replace(" ", "").lower()



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/matchapi tests...")
    test_match()


def test_match():
    int_test_handler.test_result(baseriotapi.get_match(int_test_handler.match_id))



import cassiopeia.dto.statusapi
import cassiopeia.type.core.status


def get_shards():
    """
    Get the list of server shards

    Returns:
        list<Shard>: the shards
    """
    return [cassiopeia.type.core.status.Shard(shard) for shard in cassiopeia.dto.statusapi.get_shards()]


def get_shard():
    """
    Gets the status of the current region's shard

    Returns:
        ShardStatus: the status of the current region's shard
    """
    return cassiopeia.type.core.status.ShardStatus(cassiopeia.dto.statusapi.get_shard())



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/staticdataapi tests...")
    test_champion()
    test_champions()
    test_item()
    test_items()
    test_language_strings()
    test_languages()
    test_maps()
    test_mastery()
    test_masteries()
    test_realm()
    test_rune()
    test_runes()
    test_summoner_spell()
    test_summoner_spells()
    test_versions()


def test_champion():
    int_test_handler.test_result(baseriotapi.get_champion(int_test_handler.champion_id))


def test_champions():
    int_test_handler.test_result(baseriotapi.get_champions())


def test_item():
    int_test_handler.test_result(baseriotapi.get_item(int_test_handler.item_id))


def test_items():
    int_test_handler.test_result(baseriotapi.get_items())


def test_language_strings():
    int_test_handler.test_result(baseriotapi.get_language_strings())


def test_languages():
    int_test_handler.test_result(baseriotapi.get_languages())


def test_maps():
    int_test_handler.test_result(baseriotapi.get_maps())


def test_mastery():
    int_test_handler.test_result(baseriotapi.get_mastery(int_test_handler.mastery_id))


def test_masteries():
    int_test_handler.test_result(baseriotapi.get_masteries())


def test_realm():
    int_test_handler.test_result(baseriotapi.get_realm())


def test_rune():
    int_test_handler.test_result(baseriotapi.get_rune(int_test_handler.rune_id))


def test_runes():
    int_test_handler.test_result(baseriotapi.get_runes())


def test_summoner_spell():
    int_test_handler.test_result(baseriotapi.get_summoner_spell(int_test_handler.summoner_spell_id))


def test_summoner_spells():
    int_test_handler.test_result(baseriotapi.get_summoner_spells())


def test_versions():
    int_test_handler.test_result(baseriotapi.get_versions())



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/stats tests...")
    test_ranked_stats()
    test_stats()


def test_ranked_stats():
    int_test_handler.test_result(riotapi.get_ranked_stats(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))


def test_stats():
    int_test_handler.test_result(riotapi.get_stats(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/team tests...")
    test_teams_by_summoner_id()
    test_teams_by_id()


def test_teams_by_summoner_id():
    int_test_handler.test_result(baseriotapi.get_teams_by_summoner_id(int_test_handler.summoner_id))


def test_teams_by_id():
    int_test_handler.test_result(baseriotapi.get_teams_by_id(int_test_handler.team_id))



import os

from cassiopeia import riotapi
from cassiopeia.type.core.common import LoadPolicy


RIOT_API_KEY = os.environ.get("RIOT_API_KEY")

if RIOT_API_KEY:
    riotapi.set_api_key(os.environ["RIOT_API_KEY"])
    riotapi.set_region("NA")
    riotapi.print_calls(True)
    riotapi.set_load_policy(LoadPolicy.lazy)

non_existent_champion_id = 1000
champion_id = 35
champion_name = "Thresh"
mastery_id = 6361
match_id = 1505030444
rune_id = 5234
summoner_id = 22508641
summoner_name = "FatalElement"
summoner_spell_id = 7
team_id = "TEAM-49fc9f10-1290-11e3-80a6-782bcb4d0bb2"
item_id = 3031


def test_result(result=None):
    assert True  # ??? was Pass previously



import json

try:
    import sqlalchemy
    import sqlalchemy.types
    import sqlalchemy.ext.declarative
    sqlalchemy_imported = True
except ImportError:
    sqlalchemy_imported = False


class CassiopeiaDto(object):
    """
    A Python representation of an object returned by the RiotAPI
    """

    def __init__(self, dictionary):
        """
        A Python representation of an object returned by the RiotAPI
        """
        for k, v in dictionary.items():
            setattr(self, k, v)

    def to_json(self, **kwargs):
        """
        Args:
            dictionary (dict): the JSON data returned from the Riot API as a dict
        """
        dictionary = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        default = kwargs.pop("default", lambda o: {k: v for k, v in o.__dict__.items() if not k.startswith("_")})
        sort_keys = kwargs.pop("sort_keys", True)
        indent = kwargs.pop("indent", 4)
        return json.dumps(dictionary, default=default, sort_keys=sort_keys, indent=indent, **kwargs)

    def __str__(self):
        return self.to_json()

    def __repr__(self):
        return "{class_}({dict_})".format(class_=self.__class__.__name__, dict_=self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__ if other else False

    def __ne__(self, other):
        return self.__dict__ != other.__dict__ if other else True

    def __hash__(self):
        return hash(id(self))


class CassiopeiaParametersDto(CassiopeiaDto):
    """
    Gets a JSON representation of the object

    Returns:
        str: a JSON representation of the object
    """

    def to_json(self, **kwargs):
        """
        Gets a JSON representation of the object

        Returns:
            str: a JSON representation of the object
        """
        dictionary = {k: v for k, v in self.__dict__.items() if not k.startswith("_") and v}
        default = kwargs.pop("default", lambda o: {k: v for k, v in o.__dict__.items() if not k.startswith("_") and v})
        sort_keys = kwargs.pop("sort_keys", True)
        indent = kwargs.pop("indent", 4)
        return json.dumps(dictionary, default=default, sort_keys=sort_keys, indent=indent, **kwargs)


if sqlalchemy_imported:
    BaseDB = sqlalchemy.ext.declarative.declarative_base()

    class JSONEncoded(sqlalchemy.types.TypeDecorator):
        """
        Parameters for a POST or PUT request to the Riot API
        """

        impl = sqlalchemy.Text

        def process_bind_param(self, value, dialect):
            if value is not None:
                value = json.dumps(value)
            return value

        def process_result_value(self, value, dialect):
            if value is not None:
                value = json.loads(value)
            return value



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/championapi tests...")
    test_champion_status()
    test_champion_statuses()


def test_champion_status():
    int_test_handler.test_result(riotapi.get_champion_status(riotapi.get_champion_by_id(int_test_handler.champion_id)))


def test_champion_statuses():
    int_test_handler.test_result(riotapi.get_champion_statuses())



import cassiopeia.dto.requests
import cassiopeia.type.dto.game


def get_recent_games(summoner_id):
    """
    https://developer.riotgames.com/api/methods#!/1016/3445

    Args:
        summoner_id (int): the ID of the summoner to find recent games for

    Returns:
        RecentGames: the summoner's recent games
    """
    request = "{version}/game/by-summoner/{summoner_id}/recent".format(version=cassiopeia.dto.requests.api_versions["game"], summoner_id=summoner_id)
    return cassiopeia.type.dto.game.RecentGames(cassiopeia.dto.requests.get(request))



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/matchapi tests...")
    test_match()
    test_matches()


def test_match():
    int_test_handler.test_result(riotapi.get_match(int_test_handler.match_id))


def test_matches():
    int_test_handler.test_result(riotapi.get_matches([int_test_handler.match_id]))



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class PlayerStatsSummaryList(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        playerStatSummaries (list<PlayerStatsSummary>): collection of player stats summaries associated with the summoner
        summonerId (int): summoner ID
    """
    def __init__(self, dictionary):
        self.playerStatSummaries = [(PlayerStatsSummary(pss) if not isinstance(pss, PlayerStatsSummary) else pss) for pss in dictionary.get("playerStatSummaries", []) if pss]
        self.summonerId = dictionary.get("summonerId", 0)


@cassiopeia.type.core.common.inheritdocs
class PlayerStatsSummary(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        losses (int): number of losses for this queue type. Returned for ranked queue types only
        modifyDate (int): date stats were last modified specified as epoch milliseconds
        playerStatSummaryType (str): player stats summary type (Legal values: AramUnranked5x5, Ascension, CAP5x5, CoopVsAI, CoopVsAI3x3, CounterPick, FirstBlood1x1, FirstBlood2x2, Hexakill, KingPoro, NightmareBot, OdinUnranked, OneForAll5x5, RankedPremade3x3, RankedPremade5x5, RankedSolo5x5, RankedTeam3x3, RankedTeam5x5, SummonersRift6x6, Unranked, Unranked3x3, URF, URFBots)
        wins (int): number of wins for this queue type
    """
    def __init__(self, dictionary):
        val = dictionary.get("aggregatedStats", None)
        self.aggregatedStats = None if not val else AggregatedStats(val) if not isinstance(val, AggregatedStats) else val
        self.losses = dictionary.get("losses", 0)
        self.modifyDate = dictionary.get("modifyDate", 0)
        self.playerStatSummaryType = dictionary.get("playerStatSummaryType", "")
        self.wins = dictionary.get("wins", 0)


@cassiopeia.type.core.common.inheritdocs
class AggregatedStats(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        averageAssists (int): dominion only
        averageChampionsKilled (int): dominion only
        averageCombatPlayerScore (int): dominion only
        averageNodeCapture (int): dominion only
        averageNodeCaptureAssist (int): dominion only
        averageNodeNeutralize (int): dominion only
        averageNodeNeutralizeAssist (int): dominion only
        averageNumDeaths (int): dominion only
        averageObjectivePlayerScore (int): dominion only
        averageTeamObjective (int): dominion only
        averageTotalPlayerScore (int): dominion only
        botGamesPlayed (int): botGamesPlayed
        killingSpree (int): killingSpree
        maxAssists (int): dominion only
        maxChampionsKilled (int): maxChampionsKilled
        maxCombatPlayerScore (int): dominion only
        maxLargestCriticalStrike (int): maxLargestCriticalStrike
        maxLargestKillingSpree (int): maxLargestKillingSpree
        maxNodeCapture (int): dominion only
        maxNodeCaptureAssist (int): dominion only
        maxNodeNeutralize (int): dominion only
        maxNodeNeutralizeAssist (int): dominion only
        maxNumDeaths (int): only returned for ranked statistics.
        maxObjectivePlayerScore (int): dominion only
        maxTeamObjective (int): dominion only
        maxTimePlayed (int): maxTimePlayed
        maxTimeSpentLiving (int): maxTimeSpentLiving
        maxTotalPlayerScore (int): dominion only
        mostChampionKillsPerSession (int): mostChampionKillsPerSession
        mostSpellsCast (int): mostSpellsCast
        normalGamesPlayed (int): normalGamesPlayed
        rankedPremadeGamesPlayed (int): rankedPremadeGamesPlayed
        rankedSoloGamesPlayed (int): rankedSoloGamesPlayed
        totalAssists (int): totalAssists
        totalChampionKills (int): totalChampionKills
        totalDamageDealt (int): totalDamageDealt
        totalDamageTaken (int): totalDamageTaken
        totalDeathsPerSession (int): only returned for ranked statistics
        totalDoubleKills (int): totalDoubleKills
        totalFirstBlood (int): totalFirstBlood
        totalGoldEarned (int): totalGoldEarned
        totalHeal (int): totalHeal
        totalMagicDamageDealt (int): totalMagicDamageDealt
        totalMinionKills (int): totalMinionKills
        totalNeutralMinionsKilled (int): totalNeutralMinionsKilled
        totalNodeCapture (int): dominion only
        totalNodeNeutralize (int): dominion only
        totalPentaKills (int): totalPentaKills
        totalPhysicalDamageDealt (int): totalPhysicalDamageDealt
        totalQuadraKills (int): totalQuadraKills
        totalSessionsLost (int): totalSessionsLost
        totalSessionsPlayed (int): totalSessionsPlayed
        totalSessionsWon (int): totalSessionsWon
        totalTripleKills (int): totalTripleKills
        totalTurretsKilled (int): totalTurretsKilled
        totalUnrealKills (int): totalUnrealKills
    """
    def __init__(self, dictionary):
        self.averageAssists = dictionary.get("averageAssists", 0)
        self.averageChampionsKilled = dictionary.get("averageChampionsKilled", 0)
        self.averageCombatPlayerScore = dictionary.get("averageCombatPlayerScore", 0)
        self.averageNodeCapture = dictionary.get("averageNodeCapture", 0)
        self.averageNodeCaptureAssist = dictionary.get("averageNodeCaptureAssist", 0)
        self.averageNodeNeutralize = dictionary.get("averageNodeNeutralize", 0)
        self.averageNodeNeutralizeAssist = dictionary.get("averageNodeNeutralizeAssist", 0)
        self.averageNumDeaths = dictionary.get("averageNumDeaths", 0)
        self.averageObjectivePlayerScore = dictionary.get("averageObjectivePlayerScore", 0)
        self.averageTeamObjective = dictionary.get("averageTeamObjective", 0)
        self.averageTotalPlayerScore = dictionary.get("averageTotalPlayerScore", 0)
        self.botGamesPlayed = dictionary.get("botGamesPlayed", 0)
        self.killingSpree = dictionary.get("killingSpree", 0)
        self.maxAssists = dictionary.get("maxAssists", 0)
        self.maxChampionsKilled = dictionary.get("maxChampionsKilled", 0)
        self.maxCombatPlayerScore = dictionary.get("maxCombatPlayerScore", 0)
        self.maxLargestCriticalStrike = dictionary.get("maxLargestCriticalStrike", 0)
        self.maxLargestKillingSpree = dictionary.get("maxLargestKillingSpree", 0)
        self.maxNodeCapture = dictionary.get("maxNodeCapture", 0)
        self.maxNodeCaptureAssist = dictionary.get("maxNodeCaptureAssist", 0)
        self.maxNodeNeutralize = dictionary.get("maxNodeNeutralize", 0)
        self.maxNodeNeutralizeAssist = dictionary.get("maxNodeNeutralizeAssist", 0)
        self.maxNumDeaths = dictionary.get("maxNumDeaths", 0)
        self.maxObjectivePlayerScore = dictionary.get("maxObjectivePlayerScore", 0)
        self.maxTeamObjective = dictionary.get("maxTeamObjective", 0)
        self.maxTimePlayed = dictionary.get("maxTimePlayed", 0)
        self.maxTimeSpentLiving = dictionary.get("maxTimeSpentLiving", 0)
        self.maxTotalPlayerScore = dictionary.get("maxTotalPlayerScore", 0)
        self.mostChampionKillsPerSession = dictionary.get("mostChampionKillsPerSession", 0)
        self.mostSpellsCast = dictionary.get("mostSpellsCast", 0)
        self.normalGamesPlayed = dictionary.get("normalGamesPlayed", 0)
        self.rankedPremadeGamesPlayed = dictionary.get("rankedPremadeGamesPlayed", 0)
        self.rankedSoloGamesPlayed = dictionary.get("rankedSoloGamesPlayed", 0)
        self.totalAssists = dictionary.get("totalAssists", 0)
        self.totalChampionKills = dictionary.get("totalChampionKills", 0)
        self.totalDamageDealt = dictionary.get("totalDamageDealt", 0)
        self.totalDamageTaken = dictionary.get("totalDamageTaken", 0)
        self.totalDeathsPerSession = dictionary.get("totalDeathsPerSession", 0)
        self.totalDoubleKills = dictionary.get("totalDoubleKills", 0)
        self.totalFirstBlood = dictionary.get("totalFirstBlood", 0)
        self.totalGoldEarned = dictionary.get("totalGoldEarned", 0)
        self.totalHeal = dictionary.get("totalHeal", 0)
        self.totalMagicDamageDealt = dictionary.get("totalMagicDamageDealt", 0)
        self.totalMinionKills = dictionary.get("totalMinionKills", 0)
        self.totalNeutralMinionsKilled = dictionary.get("totalNeutralMinionsKilled", 0)
        self.totalNodeCapture = dictionary.get("totalNodeCapture", 0)
        self.totalNodeNeutralize = dictionary.get("totalNodeNeutralize", 0)
        self.totalPentaKills = dictionary.get("totalPentaKills", 0)
        self.totalPhysicalDamageDealt = dictionary.get("totalPhysicalDamageDealt", 0)
        self.totalQuadraKills = dictionary.get("totalQuadraKills", 0)
        self.totalSessionsLost = dictionary.get("totalSessionsLost", 0)
        self.totalSessionsPlayed = dictionary.get("totalSessionsPlayed", 0)
        self.totalSessionsWon = dictionary.get("totalSessionsWon", 0)
        self.totalTripleKills = dictionary.get("totalTripleKills", 0)
        self.totalTurretsKilled = dictionary.get("totalTurretsKilled", 0)
        self.totalUnrealKills = dictionary.get("totalUnrealKills", 0)


@cassiopeia.type.core.common.inheritdocs
class RankedStats(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        champions (list<ChampionStats>): collection of aggregated stats summarized by champion
        modifyDate (int): date stats were last modified specified as epoch milliseconds
        summonerId (int): summoner ID
    """
    def __init__(self, dictionary):
        self.champions = [(ChampionStats(c) if not isinstance(c, ChampionStats) else c) for c in dictionary.get("champions", []) if c]
        self.modifyDate = dictionary.get("modifyDate", 0)
        self.summonerId = dictionary.get("summonerId", 0)

    @property
    def champion_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for s in self.champions:
            if s.id:
                ids.add(s.id)
        return ids


@cassiopeia.type.core.common.inheritdocs
class ChampionStats(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Gets all champion IDs contained in this object
    """
    def __init__(self, dictionary):
        self.id = dictionary.get("id", 0)
        val = dictionary.get("stats", None)
        self.stats = AggregatedStats(val) if val and not isinstance(val, AggregatedStats) else val


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_player_stats_summary():
    global PlayerStatsSummary

    @cassiopeia.type.core.common.inheritdocs
    class PlayerStatsSummary(PlayerStatsSummary, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "PlayerStatsSummary"
        aggregatedStats = sqlalchemy.orm.relationship("cassiopeia.type.dto.stats.AggregatedStats", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        losses = sqlalchemy.Column(sqlalchemy.Integer)
        modifyDate = sqlalchemy.Column(sqlalchemy.BigInteger)
        playerStatSummaryType = sqlalchemy.Column(sqlalchemy.String(30))
        wins = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)


def _sa_bind_aggregated_stats():
    global AggregatedStats

    @cassiopeia.type.core.common.inheritdocs
    class AggregatedStats(AggregatedStats, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "AggregatedStats"
        averageAssists = sqlalchemy.Column(sqlalchemy.Integer)
        averageChampionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        averageCombatPlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        averageNodeCapture = sqlalchemy.Column(sqlalchemy.Integer)
        averageNodeCaptureAssist = sqlalchemy.Column(sqlalchemy.Integer)
        averageNodeNeutralize = sqlalchemy.Column(sqlalchemy.Integer)
        averageNodeNeutralizeAssist = sqlalchemy.Column(sqlalchemy.Integer)
        averageNumDeaths = sqlalchemy.Column(sqlalchemy.Integer)
        averageObjectivePlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        averageTeamObjective = sqlalchemy.Column(sqlalchemy.Integer)
        averageTotalPlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        botGamesPlayed = sqlalchemy.Column(sqlalchemy.Integer)
        killingSpree = sqlalchemy.Column(sqlalchemy.Integer)
        maxAssists = sqlalchemy.Column(sqlalchemy.Integer)
        maxChampionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        maxCombatPlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        maxLargestCriticalStrike = sqlalchemy.Column(sqlalchemy.Integer)
        maxLargestKillingSpree = sqlalchemy.Column(sqlalchemy.Integer)
        maxNodeCapture = sqlalchemy.Column(sqlalchemy.Integer)
        maxNodeCaptureAssist = sqlalchemy.Column(sqlalchemy.Integer)
        maxNodeNeutralize = sqlalchemy.Column(sqlalchemy.Integer)
        maxNodeNeutralizeAssist = sqlalchemy.Column(sqlalchemy.Integer)
        maxNumDeaths = sqlalchemy.Column(sqlalchemy.Integer)
        maxObjectivePlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        maxTeamObjective = sqlalchemy.Column(sqlalchemy.Integer)
        maxTimePlayed = sqlalchemy.Column(sqlalchemy.Integer)
        maxTimeSpentLiving = sqlalchemy.Column(sqlalchemy.Integer)
        maxTotalPlayerScore = sqlalchemy.Column(sqlalchemy.Integer)
        mostChampionKillsPerSession = sqlalchemy.Column(sqlalchemy.Integer)
        mostSpellsCast = sqlalchemy.Column(sqlalchemy.Integer)
        normalGamesPlayed = sqlalchemy.Column(sqlalchemy.Integer)
        rankedPremadeGamesPlayed = sqlalchemy.Column(sqlalchemy.Integer)
        rankedSoloGamesPlayed = sqlalchemy.Column(sqlalchemy.Integer)
        totalAssists = sqlalchemy.Column(sqlalchemy.Integer)
        totalChampionKills = sqlalchemy.Column(sqlalchemy.Integer)
        totalDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        totalDamageTaken = sqlalchemy.Column(sqlalchemy.Integer)
        totalDeathsPerSession = sqlalchemy.Column(sqlalchemy.Integer)
        totalDoubleKills = sqlalchemy.Column(sqlalchemy.Integer)
        totalFirstBlood = sqlalchemy.Column(sqlalchemy.Integer)
        totalGoldEarned = sqlalchemy.Column(sqlalchemy.Integer)
        totalHeal = sqlalchemy.Column(sqlalchemy.Integer)
        totalMagicDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        totalMinionKills = sqlalchemy.Column(sqlalchemy.Integer)
        totalNeutralMinionsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        totalNodeCapture = sqlalchemy.Column(sqlalchemy.Integer)
        totalNodeNeutralize = sqlalchemy.Column(sqlalchemy.Integer)
        totalPentaKills = sqlalchemy.Column(sqlalchemy.Integer)
        totalPhysicalDamageDealt = sqlalchemy.Column(sqlalchemy.Integer)
        totalQuadraKills = sqlalchemy.Column(sqlalchemy.Integer)
        totalSessionsLost = sqlalchemy.Column(sqlalchemy.Integer)
        totalSessionsPlayed = sqlalchemy.Column(sqlalchemy.Integer)
        totalSessionsWon = sqlalchemy.Column(sqlalchemy.Integer)
        totalTripleKills = sqlalchemy.Column(sqlalchemy.Integer)
        totalTurretsKilled = sqlalchemy.Column(sqlalchemy.Integer)
        totalUnrealKills = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _summary_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("PlayerStatsSummary._id", ondelete="CASCADE"))


def _sa_bind_all():
    _sa_bind_player_stats_summary()
    _sa_bind_aggregated_stats()



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/featuredgamesapi tests...")
    test_featured_games()


def test_featured_games():
    int_test_handler.test_result(baseriotapi.get_featured_games())



import cassiopeia.dto.requests
import cassiopeia.type.dto.champion


def get_champion_status(id_):
    """
    https://developer.riotgames.com/api/methods#!/1015/3443

    Args:
        id (int): the ID of the champion to look up

    Returns:
        Champion: the champion
    """
    request = "{version}/champion/{id_}".format(version=cassiopeia.dto.requests.api_versions["champion"], id_=id_)
    return cassiopeia.type.dto.champion.Champion(cassiopeia.dto.requests.get(request))


def get_champion_statuses(freeToPlay=False):
    """
    https://developer.riotgames.com/api/methods#!/1015/3444

    Args:
        freeToPlay (bool): whether to only get free to play champions (default False)

    Returns:
        list<Champion>: all the champions
    """
    request = "{version}/champion".format(version=cassiopeia.dto.requests.api_versions["champion"])
    return cassiopeia.type.dto.champion.ChampionList(cassiopeia.dto.requests.get(request, {"freeToPlay": freeToPlay}))



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/gameapi tests...")
    test_recent_games()


def test_recent_games():
    int_test_handler.test_result(baseriotapi.get_recent_games(int_test_handler.summoner_id))



import os
import pytest


def pytest_runtest_setup(item):
    # Skip all these tests if we don't have RIOT_API_KEY env var
    if not os.environ.get("RIOT_API_KEY"):
        pytest.skip('No RIOT_API_KEY env var found')



import datetime

import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.team


@cassiopeia.type.core.common.inheritdocs
class Team(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.team.Team

    def __str__(self):
        return self.name

    def __iter__(self):
        return iter(self.roster)

    def __len__(self):
        return len(self.roster)

    def __getitem__(self, index):
        return self.roster[index]

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @cassiopeia.type.core.common.lazyproperty
    def creation(self):
        """
        Returns:
            datetime: the creation date of the team
        """
        return datetime.datetime.utcfromtimestamp(self.data.createDate / 1000) if self.data.createDate else None

    @property
    def id(self):
        """
        Returns:
            int: the team's id
        """
        return self.data.fullId

    @cassiopeia.type.core.common.lazyproperty
    def last_game(self):
        """
        Returns:
            datetime: the date and time for the team's last game in epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.lastGameDate / 1000) if self.data.lastGameDate else None

    @cassiopeia.type.core.common.lazyproperty
    def last_join(self):
        """
        Returns:
            datetime: the date and time for when the most recent team member joined in epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.lastJoinDate / 1000) if self.data.lastJoinDate else None

    @cassiopeia.type.core.common.lazyproperty
    def last_queue(self):
        """
        Returns:
            datetime: the date the team last joined the ranked team queue in epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.lastJoinedRankedTeamQueueDate / 1000) if self.data.lastJoinedRankedTeamQueueDate else None

    @cassiopeia.type.core.common.lazyproperty
    def match_history(self):
        """
        Returns:
            list<MatchSummary>: the match history of the team
        """
        return [MatchSummary(summary) for summary in self.data.matchHistory]

    @cassiopeia.type.core.common.lazyproperty
    def modify(self):
        """
        Returns:
            datetime: the date that team was last modified specified as epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.modifyDate / 1000) if self.data.modifyDate else None

    @property
    def name(self):
        """
        Returns:
            str: the name of the team
        """
        return self.data.name

    @property
    def captain(self):
        """
        Returns:
            Summoner: the captain of the team (returns a summoner)
        """
        return cassiopeia.riotapi.get_summoner_by_id(self.data.roster.ownerId)

    @property
    def roster(self):
        """
        Returns:
            list<TeamMember>: the team members
        """
        return [TeamMember(member) for member in self.data.roster.memberList]

    @cassiopeia.type.core.common.lazyproperty
    def second_to_last_join(self):
        """
        Returns:
            datetime: the date the second to last member joined specified as epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.secondLastJoinDate / 1000) if self.data.secondLastJoinDate else None

    @property
    def status(self):
        """
        Returns:
            str: the status of the team
        """
        return self.data.status

    @property
    def tag(self):
        """
        Returns:
            str: the team's tag
        """
        return self.data.tag

    @cassiopeia.type.core.common.lazyproperty
    def stats(self):
        """
        Returns:
            Stats: the team's stats
        """
        return [Stats(stats) for stats in self.data.teamStatDetails]

    @cassiopeia.type.core.common.lazyproperty
    def third_to_last_join(self):
        """
        Returns:
            datetime: the date the third to last member joined specified as epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.thirdLastJoinDate / 1000) if self.data.thirdLastJoinDate else None

    @cassiopeia.type.core.common.immutablemethod
    def leagues(self):
        """
        Returns:
            datetime: the creation date of the team
        """
        return cassiopeia.riotapi.get_leagues_by_team(self)

    @cassiopeia.type.core.common.immutablemethod
    def league_entries(self):
        """
        Returns:
            int: the team's id
        """
        return cassiopeia.riotapi.get_league_entries_by_team(self)


@cassiopeia.type.core.common.inheritdocs
class MatchSummary(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.team.MatchHistorySummary

    def __str__(self):
        return "Match #{id_}".format(id_=self.id)

    @property
    def kda(self):
        """
        Returns:
            float: the participant's kda
        """
        return (self.kills + self.assists) / (self.deaths if self.deaths else 1)

    @property
    def assists(self):
        """
        Returns:
            int: the number of assists the team had
        """
        return self.data.assists

    @cassiopeia.type.core.common.lazyproperty
    def date(self):
        """
        Returns:
            datetime: the date that match was completed specified as epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.date / 1000) if self.data.date else None

    @property
    def deaths(self):
        """
        Returns:
            int: the number of deaths the team had
        """
        return self.data.deaths

    @property
    def id(self):
        """
        Returns:
            datetime: the date and time for the team's last game in epoch milliseconds
        """
        return self.data.gameId

    @property
    def mode(self):
        """
        Returns:
            GameMode: the game mode of the match
        """
        return cassiopeia.type.core.common.GameMode(self.data.gameMode) if self.data.gameMode else None

    @property
    def invalid(self):
        """
        Returns:
            bool: whether or not the data is valid?
        """
        return self.data.invalid

    @property
    def kills(self):
        """
        Returns:
            int: the number of kills the team had
        """
        return self.data.kills

    @property
    def map(self):
        """
        Returns:
            Map: the map that the game was played on
        """
        return cassiopeia.type.core.common.Map(self.data.mapId) if self.data.mapId else None

    @property
    def opponent_kills(self):
        """
        Returns:
            int: the number of kills that the opponent had
        """
        return self.data.opposingTeamKills

    @property
    def opponent(self):
        """
        Returns:
            str: the name of the opposing team
        """
        return self.data.opposingTeamName

    @property
    def win(self):
        """
        Returns:
            bool: whether or not the team won this match
        """
        return self.data.win

    @cassiopeia.type.core.common.immutablemethod
    def match(self):
        """
        Returns:
            datetime: the date and time for the team's last game in epoch milliseconds
        """
        return cassiopeia.riotapi.get_match(self.id)


@cassiopeia.type.core.common.inheritdocs
class Stats(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.team.TeamStatDetail

    def __str__(self):
        return "Stats ({q})".format(q=self.queue)

    @property
    def average_games_played(self):
        """
        Returns:
            float: the average number of games played
        """
        return self.data.averageGamesPlayed

    @property
    def losses(self):
        """
        Returns:
            int: the number of times this team has lost
        """
        return self.data.losses

    @property
    def queue(self):
        """
        Returns:
            Queue: the queue type that these stats were aggregated for
        """
        return cassiopeia.type.core.common.Queue(self.data.teamStatType) if self.data.teamStatType else None

    @property
    def wins(self):
        """
        Returns:
            int: the number of times this team has won
        """
        return self.data.wins


@cassiopeia.type.core.common.inheritdocs
class TeamMember(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.team.TeamMemberInfo

    def __str__(self):
        return str(self.summoner)

    @cassiopeia.type.core.common.lazyproperty
    def invite(self):
        """
        Returns:
            datetime: the date this team member was invited to team specified as epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.inviteDate / 1000) if self.data.inviteDate else None

    @cassiopeia.type.core.common.lazyproperty
    def join(self):
        """
        Returns:
            datetime: the date this team member joined the team specified as epoch milliseconds
        """
        return datetime.datetime.utcfromtimestamp(self.data.joinDate / 1000) if self.data.joinDate else None

    @property
    def summoner(self):
        """
        Returns:
            Summoner: the summoner associated with this team member
        """
        return cassiopeia.riotapi.get_summoner_by_id(self.data.playerId)

    @property
    def status(self):
        """
        Returns:
            datetime: the date and time for when the most recent team member joined in epoch milliseconds
        """
        return self.data.status


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    Team.dto_type = cassiopeia.type.dto.team.Team
    MatchSummary.dto_type = cassiopeia.type.dto.team.MatchHistorySummary
    Stats.dto_type = cassiopeia.type.dto.team.TeamStatDetail
    TeamMember.dto_type = cassiopeia.type.dto.team.TeamMemberInfo
